from __future__ import annotations

from typing import Optional, Any, Union

import relationalai.early_access.builder as qb
from relationalai.early_access.builder.builder import RelationshipFieldRef, Field
from relationalai.early_access.dsl.core.utils import generate_stable_uuid
from relationalai.early_access.dsl.orm.utils import generate_rai_way_name


class Relationship(qb.Relationship):

    def __init__(self, model, madlib:Any, short_name:str="", fields:Optional[list[Field]]=None):
        super().__init__(madlib, short_name=short_name, model=model.qb_model(), fields=fields)
        self._dsl_model = model
        self._rel_roles = {field.name: self.__getitem__(field.name) for field in self._fields}
        self._readings[0] = RelationshipReading(self._dsl_model, madlib, self, short_name)

    def __getitem__(self, arg:Union[str, int, qb.Concept]) -> Any:
        rel_field_ref = super().__getitem__(arg)
        field_name = rel_field_ref._field_ref._name
        if hasattr(self, "_rel_roles"):
            if field_name in self._rel_roles:
                return self._rel_roles[field_name]
            else:
                raise ValueError(f"{arg} is undefined for {self._name}")
        return Role._from_field(rel_field_ref)

    def _guid(self):
        return generate_stable_uuid(str(self._id))

    def alt(self, madlib:Any, short_name:str = "", reading:qb.RelationshipReading|None = None) -> qb.RelationshipReading:
        return super().alt(madlib, short_name=short_name,
                           reading=RelationshipReading(self._dsl_model, madlib, self, short_name))

    def _unary(self):
        return self._arity() == 1

    def _binary(self):
        return self._arity() == 2

    def _first(self):
        return self.__getitem__(0)

    def _roles(self):
        return [self._rel_roles[field.name] for field in self._fields]


class RelationshipReading(qb.RelationshipReading):

    def __init__(self, model, madlib:Any, alt_of:Relationship, short_name:str):
        super().__init__(madlib, alt_of, short_name, model=model.qb_model())
        self._dsl_model = model

    def __getitem__(self, arg:Union[str, int, qb.Concept]) -> Any:
        return Role._from_field(super().__getitem__(arg))

    def _guid(self):
        return generate_stable_uuid(str(self._id))

    def _unary(self):
        return self._arity() == 1

    def _binary(self):
        return self._arity() == 2

    def _first(self):
        return self.__getitem__(0)

    def _roles(self):
        return [self._alt_of._rel_roles[field.name] for field in self._fields]

    def rai_way_name(self):
        return generate_rai_way_name(self)


class Role(RelationshipFieldRef):
    _sibling: Optional[Role] = None
    _prefix: Optional[str] = None
    _postfix: Optional[str] = None

    def __init__(self, parent:Any, part_of, pos):
        super().__init__(parent, part_of, pos)

    def _guid(self):
        return generate_stable_uuid(f"{self._field_ix}_{self._part_of()._guid()}")

    def player(self) -> qb.Concept:
        return self._concept

    def sibling(self):
        if self._relationship._arity() == 2 and not self._sibling:
            first_role = self._relationship[0]
            sibling = self._relationship[1] if self._id == first_role._id else first_role
            self._sibling = sibling
        return self._sibling
    
    def siblings(self):
        return [self._relationship[i] for i in range(self._relationship._arity()) if i != self._field_ix]

    def _part_of(self):
        return self._relationship

    def verbalization(self, prefix: Optional[str] = None, postfix: Optional[str] = None):
        self._prefix = prefix
        self._postfix = postfix

    def verbalize(self):
        text_frags = []
        if self._prefix is not None:
            text_frags.append(f"{self._prefix}-")
        text_frags.append(f"{str(self.player())}")
        if self._postfix is not None:
            text_frags.append(f"-{self._postfix}")
        return " ".join(text_frags)

    @property
    def postfix(self) -> Optional[str]:
        return self._postfix

    @property
    def prefix(self) -> Optional[str]:
        return self._prefix

    @staticmethod
    def _from_field(field:RelationshipFieldRef):
        return Role(field._parent, field._relationship, field._field_ix)

    def __hash__(self):
        return hash(f"Role({self._guid()})")

    def __eq__(self, other):
        if not isinstance(other, Role):
            return False
        return self._guid() == other._guid() and self._part_of() == other._part_of()



from __future__ import annotations
from typing import Any

from .. import builder as b
from .std import _Date, _DateTime, _Number, _String, _Integer, _make_expr
from . import dates, math, strings, decimals, integers, pragmas

def range(*args: _Integer) -> b.Expression:
    # supports range(stop), range(start, stop), range(start, stop, step)
    if len(args) == 1:
        start, stop, step = 0, args[0], 1
    elif len(args) == 2:
        start, stop, step = args[0], args[1], 1
    elif len(args) == 3:
        start, stop, step = args
    else:
        raise ValueError(f"range expects 1, 2, or 3 arguments, got {len(args)}")
    # unlike Python, Rel's range is 1..stop inclusive, so we need to subtract 1 from stop
    return _make_expr("range", cast(b.Int64, start), cast(b.Int64, stop-1), cast(b.Int64, step), b.Int64.ref("res"))

def hash(*args: Any, type=b.Hash) -> b.Expression:
    if len(args) == 0:
        raise ValueError("hash expects at least one argument")
    return _make_expr("hash", b.TupleArg(args), type.ref("res"))

def uuid_to_string(arg:_Integer) -> b.Expression:
    return _make_expr("uuid_to_string", arg, b.String.ref("res"))

def cast(type: b.Concept, arg: _Date|_DateTime|_Number|_String) -> b.Expression:
    return _make_expr("cast", b.TypeRef(type), arg, type.ref("res"))

__all__ = [
    "range",
    "dates",
    "math",
    "strings",
    "decimals",
    "integers",
    "pragmas",
]



from __future__ import annotations

from .. import builder as b
from .std import _Number, _make_expr

def abs(value: _Number) -> b.Expression:
    return _make_expr("abs", value, b.Number.ref("res"))

def natural_log(value: _Number) -> b.Expression:
    return _make_expr("natural_log", value, b.Number.ref("res"))

def sqrt(value: _Number) -> b.Expression:
    return _make_expr("sqrt", value, b.Number.ref("res"))

def maximum(left: _Number, right: _Number) -> b.Expression:
    return _make_expr("maximum", left, right, b.Number.ref("res"))

def minimum(left: _Number, right: _Number) -> b.Expression:
    return _make_expr("minimum", left, right, b.Number.ref("res"))

def isinf(value: _Number) -> b.Expression:
    return _make_expr("isinf", value)

def isnan(value: _Number) -> b.Expression:
    return _make_expr("isnan", value)

def ceil(value: _Number) -> b.Expression:
    return _make_expr("ceil", value, b.Number.ref("res"))

def floor(value: _Number) -> b.Expression:
    return _make_expr("floor", value, b.Number.ref("res"))


from __future__ import annotations
from typing import Any, Union
import textwrap
import uuid

from relationalai.early_access.metamodel.util import ordered_set
from relationalai.early_access.builder import builder as b
from relationalai.early_access.rel.executor import RelExecutor
from .common import make_name
from relationalai.experimental.solvers import Solver

_Any = Union[b.Producer, str, float, int]
_Number = Union[b.Producer, float, int]

class SolverModelPB:
    def __init__(self, model: b.Model, num_type: str):
        assert num_type in ["cont", "int"], "Invalid numerical type, must be 'cont' or 'int'"
        self._model = model # TODO can we remove? only used for _model._to_executor
        self._num_type = num_type
        self._id = next(b._global_id)
        self.variable_relationships = ordered_set()
        prefix_u = f"SolverModel_{self._id}_"
        prefix_l = f"solvermodel_{self._id}_"

        Variable = model.Concept(prefix_u + "Variable")
        self.Variable = Variable
        self.MinObjective = model.Concept(prefix_u + "MinObjective")
        self.MaxObjective = model.Concept(prefix_u + "MaxObjective")
        self.Constraint = model.Concept(prefix_u + "Constraint")

        self._model_info = {
            "num_variables": Variable,
            "num_constraints": self.Constraint,
            "num_min_objectives": self.MinObjective,
            "num_max_objectives": self.MaxObjective,
        }

        res_type = "int" if num_type == "int" else "float"
        self.result_info = model.Relationship("{key:str} has {val:str}", short_name=(prefix_l + "result_info"))
        self.point = model.Relationship(f"{{Variable}} has {{val:{res_type}}}", short_name=(prefix_l + "point"))
        self.points = model.Relationship(f"point {{i:int}} for {{Variable}} has {{val:{res_type}}}", short_name=(prefix_l + "points"))

        b.define(b.RawSource("rel", textwrap.dedent(f"""
        declare {self.MinObjective._name}
        declare {self.MaxObjective._name}
        declare {self.Constraint._name}

        declare {prefix_l}variable_name
        declare {prefix_l}minobjective_name
        declare {prefix_l}maxobjective_name
        declare {prefix_l}constraint_name
        declare {prefix_l}minobjective_serialized
        declare {prefix_l}maxobjective_serialized
        declare {prefix_l}constraint_serialized

        def {prefix_l}minobjective_printed_expr(h, s):
            rel_primitive_solverlib_print_expr({prefix_l}minobjective_serialized[h], {prefix_l}variable_name, s)

        def {prefix_l}maxobjective_printed_expr(h, s):
            rel_primitive_solverlib_print_expr({prefix_l}maxobjective_serialized[h], {prefix_l}variable_name, s)

        def {prefix_l}constraint_printed_expr(h, s):
            rel_primitive_solverlib_print_expr({prefix_l}constraint_serialized[h], {prefix_l}variable_name, s)
        """)))

    def solve_for(self, expr: b.Relationship | b.Fragment, populate: bool = True, **kwargs):
        where = []
        if isinstance(expr, b.Fragment):
            assert expr._select and len(expr._select) == 1 and expr._where, "Fragment input for `solve_for` must have exactly one select and a where clause"
            rel = expr._select[0]
            where = expr._where
        elif isinstance(expr, b.Relationship):
            rel = expr
        else:
            raise ValueError(f"Invalid expression type {type(expr)} for `solve_for`; must be a Relationship or Fragment")

        assert rel._parent, "Relationship for `solve_for` must have a parent"
        assert rel._short_name, "Relationship for `solve_for` must have a short name"
        self.variable_relationships.add(rel)

        ent = b.select(rel._parent).where(*where)
        var = self.Variable.new(entity=ent, relationship=rel._short_name)
        defs = [var]

        # handle optional variable properties
        for (key, val) in kwargs.items():
            if key == "name":
                assert isinstance(val, (_Any, list)), f"Expected {key} to be a value or list, got {type(val)}"
                defs.append(var.name(make_name(val)))
            elif key == "type":
                assert isinstance(val, str), f"Expected {key} to be a string, got {type(val)}"
                assert val in _var_types, f"Invalid variable type {val} for `solve_for`"
                ser = _make_fo_appl_with_res(_var_types[val], var)
                defs.append(self.Constraint.new(serialized=ser))
            elif key in ("lower", "upper", "fixed"):
                assert isinstance(val, _Number), f"Expected {key} to be a number, got {type(val)}"
                op = ">=" if key == "lower" else ("<=" if key == "upper" else "=")
                ser = _make_fo_appl_with_res(_fo_comparisons[op], var, val)
                defs.append(self.Constraint.new(serialized=ser))
            else:
                raise ValueError(f"Invalid keyword argument {key} for `solve_for`")

        b.define(*defs)

        if populate:
            # TODO do something different in future. maybe delete/insert into variable relationships after solve.
            # get variable values from the result point (populated by the solver)
            val = (b.Integer if self._num_type == "int" else b.Float).ref()
            b.define(rel(val)).where(self.point(var, val))
        return var

    # min/max must take a number or a Producer
    def minimize(self, expr: _Number, name = None):
        return self._obj(self.MinObjective, expr, name)

    def maximize(self, expr: _Number, name = None):
        return self._obj(self.MaxObjective, expr, name)

    def _obj(self, concept: b.Concept, expr: _Number, name):
        sym_expr = _rewrite(expr, self)
        if not sym_expr:
            # expr is not symbolic (a constant)
            # TODO should we warn if objective is constant or do further checks?
            sym_expr = _make_fo_appl_with_res(0, expr)
        elif isinstance(sym_expr, b.ConceptMember):
            # expr probably refers to a single variable, so we need to wrap it for valid protobuf
            sym_expr = _make_fo_appl_with_res(0, sym_expr)
        obj = concept.new(serialized=sym_expr)
        defs = [obj]
        if name is not None:
            defs.append(obj.name(make_name(name)))
        b.define(*defs)
        return obj

    # satisfy must take a require Fragment
    def satisfy(self, expr: b.Fragment, check: bool = False, name = None):
        assert expr._require, "Fragment input for `satisfy` must have a require clause"
        assert not expr._select and not expr._define, "Fragment input for `satisfy` must not have a select or define clause"
        if not check:
            # remove the `require` from the model roots so it is not checked
            b._remove_roots([expr])
        # TODO maybe ensure no variables in `where`s, now `.new`s?
        sym_reqs = []
        for req in expr._require:
            sym_req = _rewrite(req, self)
            assert sym_req, f"Cannot symbolify requirement {req} in `satisfy`"
            sym_reqs.append(sym_req)
        ser = b.union(*sym_reqs) if len(sym_reqs) > 1 else sym_reqs[0]
        # TODO(coey) nested select not working properly on supply_chain, so have to put the where later, for now
        # cons = self.Constraint.new(serialized=b.select(ser).where(*expr._where))
        cons = self.Constraint.new(serialized=ser)
        defs = [cons]
        if name is not None:
            defs.append(cons.name(make_name(name)))
        b.define(*defs).where(*expr._where)
        return cons

    # print counts of the number of model components
    def summarize(self):
        to_count = [
            self.Variable,
            self.MinObjective.serialized,
            self.MaxObjective.serialized,
            self.Constraint.serialized,
        ]
        counts = b.select(*[(b.count(c) | 0) for c in to_count]).to_df() # TODO(coey) do we need the |0?
        assert counts.shape == (1, len(to_count)), f"Unexpected counts shape {counts.shape}"
        (vars, min_objs, max_objs, cons) = counts.iloc[0]
        print(f"Solver model has {vars} variables, {min_objs} minimization objectives, {max_objs} maximization objectives, and {cons} constraints")
        return None

    # print the variables and components of the model in human-readable format
    def print(self, with_names: bool = False):
        print("Printing solver model.")
        # print variables
        var_df = b.select(self.Variable.name | "_").where(self.Variable).to_df()
        if var_df.empty:
            print("No variables defined in the solver model.")
            return None
        print(f"{var_df.shape[0]} variables:")
        print(var_df.to_string(index=False, header=False))

        # print components
        comps = [
            (self.MinObjective, "minimization objectives"),
            (self.MaxObjective, "maximization objectives"),
            (self.Constraint, "constraints"),
        ]
        p = b.String.ref()
        for (e, s) in comps:
            sel = [e.name | "", p] if with_names else [p]
            comp_df = b.select(*sel).where(e.printed_expr(p)).to_df()
            if not comp_df.empty:
                print(f"{comp_df.shape[0]} {s}:")
                print(comp_df.to_string(index=False, header=False))
        return None

    # solve the model given a solver and solver options
    def solve(self, solver: Solver, log_to_console: bool = False, **kwargs):
        options = kwargs
        options["version"] = 1

        # Validate options.
        for k, v in options.items():
            if not isinstance(k, str):
                raise ValueError(f"Invalid parameter key. Expected string, got {type(k)} for {k}.")
            if not isinstance(v, (int, float, str, bool)):
                raise ValueError(
                    f"Invalid parameter value. Expected string, integer, float, or boolean, got {type(v)} for {k}."
                )

        # Run the solve query and insert the extracted result.
        input_id = uuid.uuid4()
        model_uri = f"snowflake://APP_STATE.RAI_INTERNAL_STAGE/job-inputs/solver/{input_id}/model.binpb"
        sf_input_uri = f"snowflake://job-inputs/solver/{input_id}/model.binpb"
        payload: dict[str, Any] = {"solver": solver.solver_name.lower()}
        payload["options"] = options
        payload["model_uri"] = sf_input_uri

        executor = self._model._to_executor()
        assert isinstance(executor, RelExecutor)
        prefix_l = f"solvermodel_{self._id}_"

        # 1. Materialize the model and store it.
        print("export model")
        b.select(b.count(self.Variable)).to_df() # TODO(coey) weird hack to avoid uninitialized properties error
        executor.execute_raw(textwrap.dedent(f"""
        // TODO maybe only want to pass names if printing - like in old setup
        def model_relation {{
            (:variable, {self.Variable._name});
            (:variable_name, {prefix_l}variable_name);
            (:min_objective, {prefix_l}minobjective_serialized);
            (:max_objective, {prefix_l}maxobjective_serialized);
            (:constraint, {prefix_l}constraint_serialized);
        }}

        @no_diagnostics(:EXPERIMENTAL)
        def model_string {{ rel_primitive_solverlib_model_string[model_relation] }}

        ic model_not_empty("Solver model is empty.") requires not empty(model_string)

        def config[:envelope, :content_type]: "application/octet-stream"
        def config[:envelope, :payload, :data]: model_string
        def config[:envelope, :payload, :path]: "{model_uri}"
        def export {{ config }}
        """))

        # 2. Execute job and wait for completion.
        print("execute solver job")
        job_id = solver._exec_job(payload, log_to_console=log_to_console)

        # 3. Extract result.
        print("extract result")
        extract_str = textwrap.dedent(f"""
        def raw_result {{
            load_binary["snowflake://APP_STATE.RAI_INTERNAL_STAGE/job-results/{job_id}/result.binpb"]
        }}

        ic result_not_empty("Solver result is empty.") requires not empty(raw_result)

        @no_diagnostics(:EXPERIMENTAL)
        def extracted {{ rel_primitive_solverlib_extract[raw_result] }}

        def delete[:{self.result_info._name}]: {self.result_info._name}
        def delete[:{self.point._name}]: {self.point._name}
        def delete[:{self.points._name}]: {self.points._name}

        def insert(:{self.result_info._name}, key, val):
            exists((k) | string(extracted[k], val) and ::std::mirror::lower(k, key))
        """)
        if self._num_type == "int":
            extract_str += textwrap.dedent(f"""
            def insert(:{self.point._name}, var, val):
                exists((x) | extracted(:point, var, x) and
                    ::std::mirror::convert(std::mirror::typeof[Int128], x, val)
                )
            def insert(:{self.points._name}, i, var, val):
                exists((j, x) | extracted(:points, var, j, x) and
                    ::std::mirror::convert(std::mirror::typeof[Int128], j, i) and
                    ::std::mirror::convert(std::mirror::typeof[Int128], x, val)
                )
            """)
        else:
            extract_str += textwrap.dedent(f"""
            def insert(:{self.point._name}, var, val): extracted(:point, var, val)
            def insert(:{self.points._name}, i, var, val):
                exists((j) | extracted(:points, var, j, val) and
                    ::std::mirror::convert(std::mirror::typeof[Int128], j, i)
                )
            """)
        executor.execute_raw(extract_str, readonly=False)

        print("finished solve")
        return None

    # load a particular point index from `points` into `point`
    # so it is accessible from the variable relationship
    def load_point(self, i: int):
        if not isinstance(i, int) and i >= 0:
            raise ValueError(f"Expected nonnegative integer index for point, got {i}")
        executor = self._model._to_executor()
        assert isinstance(executor, RelExecutor)
        executor.execute_raw(textwrap.dedent(f"""
        def delete[:{self.point._name}]: {self.point._name}
        def insert(:{self.point._name}, var, val): {self.points._name}(int128[{i}], var, val)
        """), readonly=False)
        return None

    # print summary of the solver result
    def summarize_result(self):
        to_get = ["error", "termination_status", "solve_time_sec", "objective_value", "solver_version", "result_count"]
        k, v = b.String.ref(), b.String.ref()
        df = b.select(k, v).where(self.result_info(k, v), k.in_(to_get)).to_df()
        assert not df.empty, "No result information"
        print(df.to_string(index=False, header=False))
        return df

    # select variable names and values in the primal result point(s)
    def variable_values(self, multiple: bool = False):
        var = self.Variable.ref()
        val = (b.Integer if self._num_type == "int" else b.Float).ref()
        if multiple:
            i = b.Integer.ref()
            return b.select(i, var.name, val).where(self.points(i, var, val))
        else:
            return b.select(var.name, val).where(self.point(var, val))

    # get scalar result information after solving
    def __getattr__(self, name: str):
        df = None
        if name in self._model_info:
            df = b.select(b.count(self._model_info[name]) | 0).to_df()
        elif name in {"error", "termination_status", "solver_version", "printed_model", "solve_time_sec", "objective_value", "result_count"}:
            val = b.String.ref()
            df = b.select(val).where(self.result_info(name, val)).to_df()
        # extract scalar from df
        if df is not None:
            if not df.shape == (1, 1):
                raise ValueError(f"Expected exactly one value for {name}, but df has shape {df.shape}")
            v = df.iloc[0, 0]
            if isinstance(v, str):
                if name == "solve_time_sec":
                    return float(v)
                elif name == "objective_value":
                    return int(v) if self._num_type == "int" else float(v)
                elif name == "result_count":
                    return int(v)
            return v
        return None

# TODO maybe structure rewriting code to be more like the compiler passes rather than in one big if-else
def _rewrite(expr: Any, sm: SolverModelPB) -> Any:
    if isinstance(expr, (int, float, str)):
        return None
    elif isinstance(expr, b.Ref):
        thing = _rewrite(expr._thing, sm)
        if thing:
            return thing.ref()
        return None
    elif isinstance(expr, b.Concept):
        return None
    elif isinstance(expr, (b.Relationship, b.RelationshipRef, b.RelationshipFieldRef)):
        rel = expr if isinstance(expr, b.Relationship) else expr._relationship
        if rel in sm.variable_relationships:
            return sm.Variable(sm.Variable.ref(), entity=expr._parent, relationship=rel._short_name)
        return None
    elif isinstance(expr, b.Expression):
        op = _rewrite(expr._op, sm) # TODO what cases is this useful for?
        op_rewritten = op is not None
        params_rewritten = False
        params = []
        for p in expr._params:
            rp = _rewrite(p, sm)
            if rp:
                params_rewritten = True
                params.append(rp)
            else:
                params.append(p)
        if op_rewritten:
            assert not params_rewritten, f"Solver rewrites cannot handle expression {expr} with symbolic operator and symbolic parameters"
            return b.Expression(op, *params)
        if not params_rewritten:
            return None
        # some arguments involve solver variables, so rewrite the expression
        assert isinstance(expr._op, b.Relationship), f"Solver rewrites cannot handle expression {expr}"
        op = expr._op._name
        assert isinstance(op, str)
        if op in _fo_operators:
            return _make_fo_appl(_fo_operators[op], *params)
        elif op in _fo_comparisons:
            return _make_fo_appl_with_res(_fo_comparisons[op], *params)
        else:
            raise NotImplementedError(f"Solver rewrites cannot handle operator {op}")
    elif isinstance(expr, b.Aggregate):
        # only the last argument can be symbolic
        start_args = expr._args[:-1]
        for arg in start_args:
            assert not _rewrite(arg, sm), f"Solver rewrites cannot handle expression {expr}; only the last argument can be symbolic"
        sym_arg = _rewrite(expr._args[-1], sm)
        if not sym_arg:
            return None
        op = expr._op._name
        assert isinstance(op, str)
        if op in _ho_operators:
            appl = b.Relationship.builtins["rel_primitive_solverlib_ho_appl"]
            agg = b.Aggregate(appl, *start_args, sym_arg, _ho_operators[op])
            agg._group = expr._group
            agg._where = expr._where
            return agg
        else:
            raise NotImplementedError(f"Solver rewrites cannot handle aggregate operator {op}")
    elif isinstance(expr, b.Union):
        # return union of the symbolified expressions, if any are symbolic
        args_rewritten = False
        args = []
        for arg in expr._args:
            ra = _rewrite(arg, sm)
            if ra:
                args_rewritten = True
                args.append(ra)
            else:
                args.append(arg)
        if args_rewritten:
            return b.union(*args)
        return None
    elif isinstance(expr, b.Fragment):
        # only support selects with one item
        assert not expr._define and not expr._require and len(expr._select) == 1, "Solver rewrites only support fragments with a single select and no define or require clauses"
        sym_select = _rewrite(expr._select[0], sm)
        if sym_select:
            return b.select(sym_select).where(*expr._where)
        return None
    raise NotImplementedError(f"Solver rewrites cannot handle {expr} of type {type(expr)}")


def _make_fo_appl_with_res(op: int, *args: Any):
    return _make_fo_appl(op, *args, b.String.ref("res"))

def _make_fo_appl(op: int, *args: Any):
    assert 2 <= len(args) <= 4
    res = args[-1]
    assert isinstance(res, b.Ref)
    if res._thing != b.String:
        res = b.String.ref("res")
    appl = b.Relationship.builtins["rel_primitive_solverlib_fo_appl"]
    return b.Expression(appl, op, b.TupleArg(args[:-1]), res)


_var_types = {
    "cont": 40,
    "int": 41,
    "bin": 42,
}

_fo_operators = {
    "+": 10,
    "-": 11,
    "*": 12,
    "/": 13,
    "^": 14,
    "abs": 20,
    "exp": 21,
    "log": 22,
    "range": 50,
}

_fo_comparisons = {
    "=": 30,
    "!=": 31,
    "<=": 32,
    ">=": 33,
    "<": 34,
    ">": 35,
    "implies": 62,
}

_ho_operators = {
    "sum": 80,
    # "product":81,
    "min": 82,
    "max": 83,
    "count": 84,
    "all_different": 90,
}




import relationalai as rai
from typing import cast, Literal


def get_active_event_table() -> str:
    provider = cast(rai.clients.snowflake.Provider, rai.Provider())
    return cast(
        str, provider.sql("show parameters like 'event_table' in account;")[0]["value"]
    )


def get_events(
    seconds_ago=None,
    start=None,
    end=None,
    event_table=None,
    format: Literal["list", "pandas", "polars", "lazy"] = "pandas",
):
    provider = cast(rai.clients.snowflake.Provider, rai.Provider())
    if event_table is None:
        event_table = get_active_event_table()
    where_clauses = []
    if seconds_ago is not None:
        where_clauses.append(
            f"timestamp > current_timestamp() - interval '{seconds_ago} seconds'"
        )
    else:
        if start is not None:
            where_clauses.append(f"timestamp > '{start}'")
        if end is not None:
            where_clauses.append(f"timestamp < '{end}'")
    if where_clauses:
        where_clause = " where " + " and ".join(where_clauses)
    else:
        where_clause = ""
    return provider.sql(f"select * from {event_table}{where_clause}", format=format)



from relationalai.early_access.dsl.core.instances import Instance, ValueTypeInstance
from relationalai.early_access.dsl.core.types import Type
from relationalai.early_access.dsl.core.types.constrained import ConstrainedValueType
from relationalai.early_access.dsl.core.types.variables import NominalTypeVariable


class ValueType(ConstrainedValueType):

    def __init__(self, nm, *args):
        super().__init__(nm)
        self._types = args
        for arg in args:
            self.build_scalar_variable([arg], {})
        self.name_type_variables()

    # Add TypeVariable *tv* as a parameter to this type
    #
    def addparam(self, tv):
        self._params.append(tv)
        return tv

    def arity(self) -> int:
        return len(self._params)

    def build_scalar_variable(self, args, kwargs):
        if len(args) == 0:
            raise Exception(f"Unexpected error in ValueType {self.display()}")

        if len(kwargs) != 0:
            raise Exception(
                f"Cannot use keyword arguments when instantiating a type parameter for ValueSubtype {self.display()}")

        T = args[0]
        if len(args) == 1:

            # Then we are instantiating T to create a new TypeVariable that will parameterize
            # this ValueType (self)
            #
            tv = T.ranges_over()
            self.addparam(tv)
            return tv
        else:
            self._type_constraints.add_atomic_constraint(self.build_atom(T, args[1:]))

    def build_relation_variable(self, args, kwargs):
        if len(args) == 0:
            raise Exception(f"Unexpected error in ValueType {self.display()}")

        if len(kwargs) != 0:
            raise Exception(
                f"Cannot use keyword arguments when instantiating a type parameter for ValueSubtype {self.display()}")

        T = args[0]
        if not isinstance(T, Type):
            raise Exception("When an AtomicConstraint is asserted in context of a ValueType, the predicate must name a Type")

        if len(args) == 1:
            # Then we are instantiating T to create a new TypeVariable that will parameterize
            # this ValueType (self)
            #

            tvar = T.ranges_over() # type: ignore
            self.addparam(tvar)

            return tvar.typevar_tuple()

        else:
            # Otherwise, we expect T to behave as a type predicate that is applied to a tuple
            # (args[1:]) of BasicTypeVariables for the purpose of imposing T's constraints
            # on those variables. We implement this by creating a renaming using tuple args[1:]
            # and then applying that renaming to each constraint that T verifies, adding the
            # result to the constraints declared by self.
            #
            self._type_constraints.add_atomic_constraint(self.build_atom(T, args[1:]))

    # Generates the string used to display the opening declaration for this ValueType
    #
    def head_display(self):
        headvars = []
        for v in self._params:
            vt = v.typeof()
            headvars.append(f"{v.display()} in {vt.display()}")
        head = ", ".join(headvars)
        return f"value type {self.display()}({head})"

    def instantiate(self, *tup):

        cvars = self._params
        if len(tup) != len(cvars):
            raise Exception(
                f"ValueType {self.display()} constructor expects {len(cvars)} arguments but was instantiated with {len(tup)}")

        params = []
        for i in range(len(cvars)):
            arg = tup[i]
            tp = cvars[i].decl
            if isinstance(arg, Instance) or not tp.nominal():
                params.append(arg)
            else:
                params.append(tp.instantiate(arg))
        return ValueTypeInstance(self, params)

    def nominal(self) -> bool:
        return True

    # Return a fresh type variable that ranges over this nominal ValueType
    #
    def ranges_over(self):
        tvar = NominalTypeVariable(self)

        # For each TypeVariable that parameterizes self, duplicate
        # it to form a part of tvar
        #
        vmap = {}
        for p in self._params:
            (pprime, vmap) = p.dupe(vmap)
            tvar._parts.append(pprime)
        tvar._vmap = vmap

        # Instantiate all scalar constraints from self using vmap

        for c in self._type_constraints._sconstraints.values():
            cnew = c.revar(vmap)
            tvar._cset._sconstraints[id(cnew)] = cnew

        for c in self._type_constraints._atoms:
            cnew = c.revar(vmap)
            tvar._cset._atoms[id(cnew)] = cnew

        return tvar

    def root_nominal_type(self): return self # type: ignore

    # A non-composite ValueType can have a root unconstrained type.
    #
    def root_unconstrained_type(self):
        if self.arity() == 1:
            return self._params[0].typeof().root_unconstrained_type()
        else:
            raise Exception(f"Composite nominal type {self.display()} has no root unconstrained type")

    def __str__(self):
        return f"ValueType({self.display()})"



import os
import re
from nicegui import ui
import json

#--------------------------------------------------
# Terminal nodes
#--------------------------------------------------

TERMINAL_NODES = [
    "query"
]

#--------------------------------------------------
# debug.jsonl helpers
#--------------------------------------------------

class SpanNode:
    def __init__(self, id, type, parent_id, start_timestamp, attrs=None):
        self.id = id
        self.type = type
        self.parent_id = parent_id
        self.start_timestamp = start_timestamp
        self.end_timestamp = None
        self.elapsed = None
        self.start_attrs = attrs or {}
        self.end_attrs = {}
        self.children = []

    def add_end_data(self, end_timestamp, elapsed, end_attrs=None):
        self.end_timestamp = end_timestamp
        self.elapsed = elapsed
        if end_attrs:
            # Merge end_attrs into attrs
            self.end_attrs.update(end_attrs)

    def add_child(self, child):
        self.children.append(child)

    def __str__(self, level=0):
        indent = "  " * level
        result = f"{indent}{self.type} ({self.id}):\n"
        result += f"{indent}  elapsed: {self.elapsed:.6f}s\n"

        if self.start_attrs or self.end_attrs:
            result += f"{indent}  start attributes:\n"
            for key, value in self.start_attrs.items():
                # Truncate long values for display
                if isinstance(value, str) and len(value) > 100:
                    value = value[:97] + "..."
                result += f"{indent}    {key}: {value}\n"
            result += f"{indent}  end attributes:\n"
            for key, value in self.end_attrs.items():
                # Truncate long values for display
                if isinstance(value, str) and len(value) > 100:
                    value = value[:97] + "..."
                result += f"{indent}    {key}: {value}\n"

        if self.children:
            result += f"{indent}  children:\n"
            for child in self.children:
                result += child.__str__(level + 2)

        return result

def parse_jsonl_to_tree(jsonl_content):
    lines = jsonl_content.strip().split('\n')
    nodes_by_id = {}
    root_nodes = []

    for line in lines:
        data = json.loads(line)
        event_type = data["event"]

        if event_type == "span_start":
            span = data["span"]
            node = SpanNode(
                id=span["id"],
                type=span["type"],
                parent_id=span["parent_id"],
                start_timestamp=span["start_timestamp"],
                attrs=span.get("attrs", {})
            )
            nodes_by_id[node.id] = node

            # Link to parent if exists
            if node.parent_id and node.parent_id in nodes_by_id:
                nodes_by_id[node.parent_id].add_child(node)
            elif node.parent_id is None:
                root_nodes.append(node)

        elif event_type == "span_end":
            node_id = data["id"]
            if node_id in nodes_by_id:
                nodes_by_id[node_id].add_end_data(
                    end_timestamp=data["end_timestamp"],
                    elapsed=data.get("elapsed", 0),
                    end_attrs=data.get("end_attrs", {})
                )

    return root_nodes

#--------------------------------------------------
# UI
#--------------------------------------------------

last_mod_time = None
current_json_objects = []
active_ix = None
active_item = None

toggles = {
    "dsl": False,
    "metamodel": False,
    "rel": False,
    "lqp": False,
    "sql": False,
    "passes": False,
    "type_graph": False,
    "expand_all": False
}

checkboxes = {}

opened = set()

def toggle_open(source):
    global opened
    if source in opened:
        opened.remove(source)
    else:
        opened.add(source)
    sidebar.refresh()
    details.refresh()

def to_ms(t):
    if not t:
        return "..."
    return f"{t*1000:,.0f}"

def format_time(t):
    if not t:
        return "..."
    if t > 1:
        return f"{t:.1f}s"
    elif t > 0.001:
        return f"{t*1000:.1f}ms"
    elif t > 0.0005:
        return f"{t*1000:.2f}ms"
    else:
        return f"{t*1000000:.0f}us"

def header(text):
    return ui.label(text).style("font-size: 1.3em; font-weight: bold;")

def replace_long_brace_contents(code_str):
    def replacement(match):
        string = match.group(0)
        if len(string) > 300:
            # Extract the first 50 and last 30 characters from the string within brackets
            return '{' + string[1:51] + '...' + string[-31:-1] + '}'
        else:
            # If the string is not longer than 300 characters, return it unchanged
            return string

    # This regex matches sequences of characters wrapped in { }
    brace_content_regex = r'\{[\s\S]*?\}'

    # Use the sub method to replace the matched strings with the result of the replacement function
    return re.sub(brace_content_regex, replacement, code_str)


def code(c, language="python"):
    # do not wrap text when outputting Rel
    if language != "rel":
        c = replace_long_brace_contents(c)
    # use python's syntax highlighting for Rel as well
    if language == "rel":
        language = "ruby"
    if language == "lqp":
        language = "clojure"
    c = re.sub(r"→", "->", c)
    c = re.sub(r"⇑", "^", c)
    return ui.code(c, language=language).style("border:none; margin:0; padding-right: 30px; ").classes("w-full")

def code_with_header(header, c, language="python"):
    header = f"# --- {header} {'-' * (40 - len(header))}"
    c = header + "\n\n" + c.strip()
    return code(c, language=language)

def span_code(span:SpanNode):
    if span.type == "rule":
        if span.start_attrs.get("source"):
            code(span.start_attrs["source"]).on("click", lambda: toggle_open(span.start_attrs["source"])).style("cursor: pointer;")
        if not toggles["expand_all"] and span.start_attrs["source"] not in opened:
            return
        with ui.column().style("gap:0px; padding-left: 30px; "):
            if toggles["dsl"]:
                code_with_header("dsl", span.start_attrs["dsl"])
            if toggles["metamodel"]:
                code_with_header("metamodel", span.end_attrs["metamodel"])
    elif span.type == "query":
        if span.start_attrs.get("source"):
            code(span.start_attrs["source"]).on("click", lambda: toggle_open(span.start_attrs["source"])).style("cursor: pointer;")
        if not toggles["expand_all"] and span.start_attrs["source"] not in opened:
            return
        with ui.column().style("gap:0px; padding-left: 30px; "):
            if toggles["dsl"]:
                code_with_header("dsl", span.start_attrs["dsl"])
            for child in span.children:
                span_code(child)
    elif span.type == "compile" and span.end_attrs.get("compile_type") == "model":
        if toggles["metamodel"]:
            code_with_header("model metamodel", span.start_attrs["metamodel"])
        for child in span.children:
            span_code(child)
        if toggles["rel"]:
            code_with_header("rel model", span.end_attrs.get("rel", '// No Rel was emitted.'), language="rel")
        if toggles["lqp"]:
            code_with_header("lqp model", span.end_attrs.get("lqp", ';; No LQP was emitted.'), language="lqp")
        if toggles["sql"]:
            code_with_header("sql model", span.end_attrs.get("sql", '-- No SQL was emitted.'), language="sql")
    elif span.type == "compile" and span.end_attrs.get("compile_type") == "query":
        if toggles["metamodel"]:
            code_with_header("metamodel", span.start_attrs["metamodel"])
        for child in span.children:
            span_code(child)
        if toggles["rel"]:
            code_with_header("rel query", span.end_attrs.get("rel", '// No Rel was emitted.'), language="rel")
        if toggles["lqp"]:
            code_with_header("lqp query", span.end_attrs.get("lqp", ';; No LQP was emitted.'), language="lqp")
        if toggles["sql"]:
            code_with_header("sql query", span.end_attrs.get("sql", '-- No SQL was emitted.'), language="sql")
    elif span.type == "passes":
        type_graph_shown = False
        if toggles["passes"]:
            previous_code = None
            for child in span.children:
                current_code = child.end_attrs["metamodel"]
                if current_code == previous_code:
                    current_code = ""
                else:
                    previous_code = current_code
                code_element = code_with_header(child.type + f' ({format_time(child.elapsed)})', current_code)
                code_element.props(f'id="pass-{child.id}"')
                code_element.style("scroll-margin-top: 40px;")
                if child.type == "InferTypes" and toggles["type_graph"]:
                    # show the type graph inline with InferTypes
                    type_graph_shown = True
                    for child2 in child.children:
                        if child2.end_attrs.get("type_graph"):
                            ui.mermaid(child2.end_attrs['type_graph'], {"maxEdges":10000}).style("display:flex; padding:0px 0px 0px 15px; gap:30px; overflow: auto;")
        # show the type graph if it was not shown inline above
        if toggles["type_graph"] and not type_graph_shown:
            for child in span.children:
                if child.type == "InferTypes":
                    for child2 in child.children:
                        if child2.end_attrs.get("type_graph"):
                            ui.mermaid(child2.end_attrs['type_graph'], {"maxEdges":10000}).style("display:flex; padding:0px 0px 0px 15px; gap:30px; overflow: auto;")
    else:
        for child in span.children:
            span_code(child)

@ui.refreshable
def details():
    with ui.column().style("gap:2px; margin-left: 230px;"):
        for root in current_json_objects:
            span_code(root)


def handle_attributes(attrs):
    if attrs.get("file"):
        ui.label(f"{attrs['file']}: {attrs['line']}")
    if attrs.get("source"):
        code(attrs["source"])
    if attrs.get("txn_id"):
        ui.label(f"txn_id: {attrs['txn_id']}")
    if attrs.get("name"):
        ui.label(f"{attrs['name']}")
    if attrs.get("code"):
        code(attrs["code"])
    if attrs.get("dsl"):
        code(attrs["dsl"])
    if attrs.get("metamodel"):
        code(attrs["metamodel"])
    if attrs.get("rel"):
        code(attrs["rel"], language="rel")
    if attrs.get("sql"):
        code(attrs["sql"], language="sql")
    if attrs.get("results"):
        vals = attrs['results']
        if len(vals):
            keys = [k for k in vals[0].keys()]
            columns = [{'name': k, 'label': k, 'field': k, "align": "left"} for k in keys]
            ui.table(columns=columns, rows=vals)

def handle_body(span: SpanNode):
    if span.children and span.type not in ["rule_batch"]:
        for child in span.children:
            span_ui(child, True)

def span_ui(span: SpanNode, is_pass: bool=False):
    def jump_to_pass():
        global toggles
        checkboxes["passes"].set_value(True)
        toggles["passes"] = True
        sidebar.refresh()
        details.refresh()
        ui.navigate.to(f"#pass-{span.id}")

    with ui.column().classes("w-full").style("padding:0px 0px 0px 15px; gap:2px; "):
        with ui.row().classes("w-full").style("padding-left:5px; flex-wrap:nowrap; margin:0; background: #2c2c2c") as me:
            name = span.type
            if name == "create_v2":
                name += " ➚"
            ui.label(name).style("padding:0; margin:0; color: #DDD;")
            ui.space()
            ui.label(f"{to_ms(span.elapsed)}").style("padding-right: 5px; color:#999;")
            if span.type == "create_v2":
                txn_id = span.end_attrs.get("txn_id")
                me.on("click", lambda: ui.navigate.to(f"https://171608476159.observeinc.com/workspace/41759331/log-explorer?datasetId=41832558&filter-rai_transaction_id={txn_id}", new_tab=True))
                me.style("cursor: pointer;")
            elif is_pass:
                me.on("click", lambda: jump_to_pass())
                me.style("cursor: pointer;")
        handle_body(span)


@ui.refreshable
def sidebar():
    with ui.column().style("min-width: 270px; max-width: 270px; margin-top: 50px; height: calc(100vh - 60px); overflow-x: hidden; overflow-y: auto; position: fixed; top: 0; left: 0; z-index: 1000;"):
        for root in current_json_objects:
            span_ui(root)


def poll():
    global last_mod_time, active_item
    global current_json_objects
    # Check the last modification time of the file
    try:
        mod_time = os.path.getmtime('debug.jsonl')
        if last_mod_time is None or mod_time > last_mod_time:
            last_mod_time = mod_time
            # File has changed, read and parse the new content
            with open('debug.jsonl', 'r') as file:
                content = file.read()
                if content:
                    new_tree = parse_jsonl_to_tree(content)
                    # Update the current JSON objects
                    current_json_objects = new_tree

                    if active_ix is not None and len(current_json_objects) > active_ix:
                        active_item = current_json_objects[active_ix]
                    # Refresh the UI with the new objects
                    sidebar.refresh()
                    details.refresh()
    except FileNotFoundError:
        pass

def toggle(key):
    global toggles
    toggles[key] = not toggles[key]
    sidebar.refresh()
    details.refresh()

def main(host="0.0.0.0", port=8080):
    global checkboxes
    ui.dark_mode(None)
    with ui.column():
        with ui.row().style("flex-wrap: nowrap; position: fixed; top: 0; left: 0; right: 0; z-index: 1000; background: #3c3c3c; color: #DDD;"):
            ui.switch('expand all', value=toggles["expand_all"], on_change=lambda: toggle("expand_all"))
            checkboxes["dsl"] = ui.checkbox('dsl', on_change=lambda: toggle("dsl"))
            checkboxes["metamodel"] = ui.checkbox('metamodel', on_change=lambda: toggle("metamodel"))
            checkboxes["passes"] = ui.checkbox('passes', on_change=lambda: toggle("passes")).props('id=passes_box')
            checkboxes["rel"] = ui.checkbox('rel', on_change=lambda: toggle("rel"))
            checkboxes["lqp"] = ui.checkbox('lqp', on_change=lambda: toggle("lqp"))
            checkboxes["sql"] = ui.checkbox('sql', on_change=lambda: toggle("sql"))
            checkboxes["type graph"] = ui.checkbox('type graph', on_change=lambda: toggle("type_graph"))
        with ui.row().style("flex-wrap: nowrap; margin-top: 30px;"):
            with ui.column() as c:
                sidebar()
            with ui.column() as c:
                c.style("padding-left: 2em;")
                details()

        # Scroll to top button
        ui.button("↑", on_click=lambda: ui.run_javascript("window.scrollTo({top: 0, behavior: 'smooth'})")).style(
            "position: fixed; "
            "bottom: 20px; "
            "right: 20px; "
            "z-index: 1001; "
            "width: 50px; "
            "height: 50px; "
            "border-radius: 50%; "
            "background: #4a4a4a; "
            "color: white; "
            "border: 2px solid #666; "
            "font-size: 20px; "
            "font-weight: bold; "
            "cursor: pointer; "
            "box-shadow: 0 2px 10px rgba(0,0,0,0.3);"
        ).props("title='Scroll to top'")

    ui.timer(1, poll)
    ui.run(reload=False, host=host, port=port)

if __name__ in {"__main__", "__mp_main__"}:
    main()



"""
Rewrite list types to tuple or scalar types depending on how they are used.
"""
from dataclasses import dataclass, field
from typing import List, cast

from relationalai.early_access.metamodel import ir, visitor, compiler, builtins as bt
from relationalai.early_access.metamodel.util import ordered_set

@dataclass
class RewriteListTypes(compiler.Pass):
    """
    Rewrite list types to tuple or scalar types depending on how they are used.
    """
    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        v = RewriteListTypesVisitor()
        result = v.walk(model)
        return result

@dataclass
class RewriteListTypesVisitor(visitor.Rewriter):
    """
    A pass that fixes the types of nodes that use ListTypes.
    """
    new_relations: List[ir.Relation] = field(default_factory=list, init=False)
    new_types: List[ir.Type] = field(default_factory=list, init=False)

    def handle_model(self, model: ir.Model, parent: None):
        result = super().handle_model(model, parent)
        relations_without_list_type = ordered_set(*[
            r for r in result.relations if not any(isinstance(f.type, ir.ListType) for f in r.fields)
        ]).frozen()
        return model.reconstruct(
            result.engines,
            relations_without_list_type | self.new_relations,
            result.types | self.new_types,
            result.root,
            result.annotations,
        )

    def handle_lookup(self, node: ir.Lookup, parent: ir.Node):
        if any(isinstance(f.type, ir.ListType) for f in node.relation.fields):
            return self._rewrite_non_aggr_relation(node, node.relation)
        else:
            return node

    def handle_aggregate(self, node: ir.Aggregate, parent: ir.Node):
        new_aggr = self._rewrite_aggr_relation(node, node.aggregation)
        return ir.Aggregate(
            node.engine,
            new_aggr,
            node.projection,
            node.group,
            node.args,
        )

    def _rewrite_aggr_relation(self, node: ir.Aggregate, relation: ir.Relation) -> ir.Relation:
        assert len(relation.fields) >= 1
        changed = False

        if changed:
            overloads = ordered_set(*[
                self._rewrite_aggr_relation(node, overload)
                for overload in relation.overloads
            ]).frozen()

            new_relation = ir.Relation(
                relation.name,
                tuple(relation.fields),
                relation.requires,
                relation.annotations,
                overloads,
            )
            self.new_relations.append(new_relation)
            return new_relation

        return relation

    def _rewrite_non_aggr_relation(self, node: ir.Lookup, relation: ir.Relation):
        # Handle the simple varargs case: one list field and all the rest are scalar.
        # Currently this pattern is only used for `rel_primitive_hash_tuple`.
        # and `rel_primitive_solverlib_fo_appl`.
        list_field_indexes = [i for i in range(len(node.relation.fields)) if isinstance(node.relation.fields[i].type, ir.ListType)]
        if len(list_field_indexes) != 1:
            return node

        # There exactly one list field, rewrite it to a tuple type.
        i = list_field_indexes[0]
        assert isinstance(node.relation.fields[i].type, ir.ListType)
        scalar_field_count = len(node.relation.fields) - 1
        tuple_len = len(node.args) - scalar_field_count
        assert tuple_len >= 0, f"List field {i} has {tuple_len} elements, but there are only {scalar_field_count} scalar fields."

        # Flatten the list field into separate scalar fields.
        field = node.relation.fields[i]
        ft = cast(ir.ListType, field.type)
        new_fields = (
            list(node.relation.fields[0:i]) +
            [ir.Field(f"{field.name}@{j}", ft.element_type, field.input) for j in range(tuple_len)] +
            list(node.relation.fields[i+1:])
        )

        # Create new relation, adding external annotation to prevent renaming.
        # NOTE(coey): maybe this is a misuse of this annotation and there could be a nicer way to do it with overloads.
        annos = ordered_set(*node.relation.annotations, bt.external_annotation).frozen()
        new_relation = ir.Relation(node.relation.name, tuple(new_fields), node.relation.requires, annos)
        self.new_relations.append(new_relation)

        return ir.Lookup(node.engine, new_relation, node.args)



# This script runs a QB program and captures high-level timing information about different parts of the system.
# The output is a JSON file containing timing information that can be used to analyze performance.

import argparse
import os
import json

from relationalai.clients.snowflake import Resources as snowflake_api
from relationalai.early_access.lqp.executor import LQPExecutor
from relationalai.early_access.builder import builder
from relationalai.clients.use_index_poller import UseIndexPoller as index_poller

from enum import Enum

import time

# Categories that spans can belong to
class SpanCategory(Enum):
    DATA_PREP = "Data Preparation"
    QB_COMPILE = "QB Compilation"
    EXEC = "RAICode Execution"
    SNOWFLAKE = "Snowflake Requests"
    RESULTS = "Result Processing"
    OTHER = "Other"
    UNKNOWN = "Unknown"

# Contextual information that describes an instrumented method. These are used to create spans.
#
# To better distinguish important spans, we have an `is_core` flag to indicate "core methods".
# Core methods represent high-level concepts that make sense to be reasoned about in isolation.
#
# Core methods must be mutually exclusive. This means that they should not overlap in time with one another,
# and should represent distinct concepts.
# For example, `exec_rai_app` and `process_results` are core methods since they are distinct high-level concepts
# that aid with understanding the overall time breakdown and do not overlap with one another.
# On the other hand, `exec_snowflake` is not a core method since it overlaps with other core methods and
# is tricky to reason about in isolation without knowing /what/ was being executed.
#
# Ideally, the sum of all core methods should equal the total time of the root span.
#
# Additionally, we have a `count_for_cat` flag to indicate whether this method should count towards the time of
# its category. This prevents duplicate counting for overlapping methods in the same category.
class TimeMarker:
    def __init__(self, name, category: SpanCategory, is_core=False, count_for_cat=False):
        self.name = name
        self.category = category
        self.is_core = is_core
        self.count_for_cat = count_for_cat

# Time span for a specific operation.
class TimeSpan:
    def __init__(self, id: str, marker: TimeMarker, parent_id, tags={}):
        self.id = id
        self.marker = marker
        self.parent_id = parent_id
        self.tags = tags

        self.start_time = None
        self.end_time = None

    def start(self):
        self.start_time = time.time()

    def end(self):
        self.end_time = time.time()

    def duration(self):
        assert self.start_time is not None, "Span has not started yet"
        assert self.end_time is not None, "Span has not ended yet"
        return self.end_time - self.start_time

# Context for the benchmarking. Keeps track of active spans and finished spans.
class BenchmarkContext:
    def __init__(self):
        self.current_spans = []
        self.finished_spans = []
        self.span_id_counter = 0
        self.print_logs = True
        self.gap_size = 0.0
        self.last_end_time = 0

    def start_span(self, marker: TimeMarker):
        parent_id = self.current_spans[-1].id if self.current_spans else None
        span_id = f"span_{self.span_id_counter}"
        self.span_id_counter += 1

        new_span = TimeSpan(span_id, marker, parent_id)
        new_span.start()
        self.current_spans.append(new_span)

        if new_span.marker.is_core:
            assert new_span.start_time is not None, "Span has not started yet"
            last_end_diff = new_span.start_time - self.last_end_time
            self.gap_size += last_end_diff
            if last_end_diff > 0.01:
                print(f"\tWarning: Gap of {last_end_diff:.4f} seconds before start of span `{new_span.id}` ({new_span.marker.name})")

        return new_span

    def end_span(self):
        assert self.current_spans, "No active span to end"
        span = self.current_spans.pop()
        span.end()
        self.finished_spans.append(span)
        if span.marker.is_core:
            self.last_end_time = span.end_time

        if self.print_logs:
            duration = span.duration()
            print(f"Span {span.id} ({span.marker.name}) ended, duration: {duration:.4f} seconds")
        return span

BENCHMARK_CTX = BenchmarkContext()

# Wrapper for instrumenting a method.
# If more fine-grained control is needed, consider monkey-patching manually with spans instead.
# E.g. if you want to add tags or have more complex logic.
class InstrumentedMethod:
    def __init__(self, module, method_name, category, is_core=False, count_for_cat=False, alias=None):
        self.module = module
        self.method_name = method_name

        marker_name = alias if alias else method_name
        self.marker = TimeMarker(marker_name, category, is_core, count_for_cat)

        self.original_method = getattr(module, method_name)
        self.new_method = self._create_instrumented_method()

    def _create_instrumented_method(self):
        def instrumented_method(*args, **kwargs):
            global BENCHMARK_CTX
            BENCHMARK_CTX.start_span(self.marker)
            result = self.original_method(*args, **kwargs)
            BENCHMARK_CTX.end_span()
            return result
        return instrumented_method

    def apply(self):
        setattr(self.module, self.method_name, self.new_method)

    def restore(self):
        setattr(self.module, self.method_name, self.original_method)

# List of all methods to track.
INSTRUMENTED_METHODS = [
    # Core methods. They must be mutually exclusive, high-level concepts, and ideally sum up to the total time.
    InstrumentedMethod(LQPExecutor, "prepare_data", SpanCategory.DATA_PREP, is_core=True, count_for_cat=True),
    InstrumentedMethod(LQPExecutor, "compile_lqp", SpanCategory.QB_COMPILE, is_core=True, count_for_cat=True),
    InstrumentedMethod(snowflake_api, "_exec_rai_app", SpanCategory.EXEC, is_core=True, count_for_cat=True, alias="exec_rai_app"),
    InstrumentedMethod(snowflake_api, "_fetch_exec_async_artifacts", SpanCategory.RESULTS, is_core=True, count_for_cat=True, alias="fetch_artifacts"),
    InstrumentedMethod(LQPExecutor, "_process_results", SpanCategory.RESULTS, is_core=True, count_for_cat=True, alias="process_results"),

    # Other trackers, these should also be on the high-level side but can overlap
    InstrumentedMethod(LQPExecutor, "execute", SpanCategory.EXEC, alias="execute_lqp"),
    InstrumentedMethod(snowflake_api, "_exec", SpanCategory.SNOWFLAKE, count_for_cat=True, alias="exec_snowflake"),
    InstrumentedMethod(index_poller, "_poll_loop", SpanCategory.DATA_PREP, alias="poll_index"),
]

def _make_unknown_span(parent_id, parent_duration, known_children_duration):
    unknown_duration = parent_duration - known_children_duration
    if unknown_duration <= 0.01:
        return None
    return {
        "name": "Unknown",
        "duration": unknown_duration,
        "category": SpanCategory.UNKNOWN.value,
        "parent": parent_id,
        "tags": {},
        "is_core": True,
        "children": [],
        "total_percent": (unknown_duration / parent_duration) * 100,
        "parent_percent": (unknown_duration / parent_duration) * 100,
    }

def _make_task_times_dict(spans, root_span):
    total_duration = root_span.duration()
    task_times = dict()
    for span in spans:
        if span.marker.name not in task_times:
            task_times[span.marker.name] = {"duration": 0.0, "category": span.marker.category.value}
        task_times[span.marker.name]["duration"] += span.duration()

    # Add total percent to each
    for v in task_times.values():
        v["percent"] = (v["duration"] / total_duration) * 100
    return task_times

def _make_spans_dict(spans, root_span):
    spans_dict = dict()
    for span in spans:
        spans_dict[span.id] = {
            "name": span.marker.name,
            "duration": span.duration(),
            "category": span.marker.category.value,
            "parent": span.parent_id,
            "tags": span.tags,
            "is_core": span.marker.is_core,

            # To be filled later
            "children": [],
        }

    # Fill in children
    for span in spans:
        if span.parent_id:
            spans_dict[span.parent_id]["children"].append(span.id)

    # Add total_percent and parent_percent
    total_duration = root_span.duration()
    for span in spans:
        span_info = spans_dict[span.id]
        span_info["total_percent"] = (span_info["duration"] / total_duration) * 100
        if span.parent_id:
            parent_duration = spans_dict[span.parent_id]["duration"]
            span_info["parent_percent"] = (span_info["duration"] / parent_duration) * 100
        else:
            # Root span has no parent
            span_info["parent_percent"] = 100.0

    # Add "Unknown" spans for when sum of children < parent
    for span in spans:
        children = spans_dict[span.id]["children"]
        if not children:
            continue
        known_children_duration = sum(spans_dict[child_id]["duration"] for child_id in children)
        unknown_span = _make_unknown_span(span.id, spans_dict[span.id]["duration"], known_children_duration)
        if unknown_span:
            unknown_id = f"{span.id}_unknown"
            spans_dict[unknown_id] = unknown_span
            spans_dict[span.id]["children"].append(unknown_id)

    return spans_dict

def _make_core_times_dict(spans, root_span):
    core_times = dict()
    total_duration = root_span.duration()
    for span in spans:
        if span.marker.is_core:
            if span.marker.name not in core_times:
                core_times[span.marker.name] = {
                    "duration": 0.0,
                }
            core_times[span.marker.name]["duration"] += span.duration()
    for v in core_times.values():
        v["total_percent"] = (v["duration"] / total_duration) * 100
    return core_times

def _make_category_times_dict(spans, root_span):
    category_times = dict()
    for span in spans:
        if span.marker.count_for_cat:
            cat_name = span.marker.category.value
            if cat_name not in category_times:
                category_times[cat_name] = {
                    "duration": 0.0,
                }
            category_times[cat_name]["duration"] += span.duration()
    return category_times

def main():
    parser = argparse.ArgumentParser(description="Extract LQP requests to run locally")
    parser.add_argument("--program", type=str, help="Path to the QB program to run", required=True)
    parser.add_argument("--outdir", type=str, help="Path to the output directory", required=True)

    args = parser.parse_args()
    program = args.program
    outdir = args.outdir

    if not os.path.isfile(program):
        raise ValueError(f"Program path {program} is not a file")
    if not os.path.isdir(outdir):
        raise ValueError(f"Output path {outdir} is not a directory")

    benchmark_lqp(program, outdir)

def _run_instrumented_qb(qb_prog_path: str):
    global BENCHMARK_CTX
    global INSTRUMENTED_METHODS

    # Run the QB program with instrumentation
    for method in INSTRUMENTED_METHODS:
        method.apply()

    # Special marker for the root span
    e2e_marker = TimeMarker("e2e", SpanCategory.OTHER)
    root_span = BENCHMARK_CTX.start_span(e2e_marker)
    BENCHMARK_CTX.last_end_time = root_span.start_time
    with builder.with_overrides(**{
        'dry_run': False,
        'reasoner.rule.use_lqp': True,
    }):
        with open(qb_prog_path) as f:
            # Compile with the correct filename
            code = compile(f.read(), qb_prog_path, "exec")
            exec(code)
    root_span = BENCHMARK_CTX.end_span()
    assert root_span.marker == e2e_marker, "Mismatched root span"
    for method in INSTRUMENTED_METHODS:
        method.restore()
    spans = BENCHMARK_CTX.finished_spans
    return spans, root_span

def benchmark_lqp(qb_prog_path: str, outdir: str):
    global BENCHMARK_CTX
    spans, root_span = _run_instrumented_qb(qb_prog_path)

    print(f"Total benchmark time: {root_span.duration():.4f} seconds")
    print(f"\tMissing time not tracked by core methods: {BENCHMARK_CTX.gap_size:.4f} seconds")

    # Write results to file
    final_json_dict = dict()
    final_json_dict["spans"] = _make_spans_dict(spans, root_span)
    final_json_dict["task_times"] = _make_task_times_dict(spans, root_span)
    final_json_dict["core_times"] = _make_core_times_dict(spans, root_span)
    final_json_dict["category_times"] = _make_category_times_dict(spans, root_span)

    final_json = json.dumps(final_json_dict, default=str)
    out_path = os.path.join(outdir, "benchmark_results.json")
    with open(out_path, "w") as f:
        f.write(final_json)

    print(f"Wrote benchmark results to {out_path}")

if __name__ == "__main__":
    main()



# Copyright 2024 RelationalAI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This module declares a simple propositional constraint language that admits
# scalar, atomic, and predicate constraints
#
# These are used to represent both Schema- and ValueType- constraints, the
# differences being what counts for a leaf variable and the expressive power.
from abc import abstractmethod


class Constraint:

    # Returns an integer (hash) that identifies this Expr, allowing us
    # to treat Exprs sort of like entity types for which we had declared
    # a preferred identifier so as to allow different physical objects
    # to unify to the same entity. Contrast this with the "id" method
    # which would not allow us to do this unification.
    #
    @abstractmethod
    def entityid(self) -> int: pass


class TrueConstraint(Constraint):

    def __init__(self): pass

    def pprint(self): return 'true'

    def refersto(self, varname: str): return False

    def rename(self, renaming): return self

    def entityid(self):
        return hash(TrueConstraint)


class FalseConstraint(Constraint):
    def __init__(self): pass

    def pprint(self): return 'false'

    def refersto(self, varname: str): return False

    def rename(self, renaming): return self

    def entityid(self):
        return hash(TrueConstraint)


# We represent "constraint sets" as dictionaries that are keyed by the entityid
# of each constraint in the set.
#

# Returns the set difference of two constraint sets C1 and C2
#
def diffof(C1, C2):
    return {c: C1[c] for c in C1 if c not in C2}


# Returns the disjoint union of two constraint sets C1 and C2
#
def disjoint_union_of(C1, C2):
    C = C1.copy()
    for c in C2:
        if c in C1:
            raise Exception(f"Common constraint {c.pprint()} found in a disjoint union of two sets")
        C[c] = C2[c]
    return C



import typing
from abc import abstractmethod, ABC
from collections import OrderedDict
from collections.abc import Iterable
from enum import Enum
from typing import Optional, Union, Callable

import relationalai.early_access.builder as qb
from relationalai.early_access.builder import define, where, not_
from relationalai.early_access.builder.builder import Not, ConceptConstruct, Ref, RelationshipRef
from relationalai.early_access.dsl.bindings.common import Binding, BindableAttribute, IdentifierConceptBinding, \
    SubtypeConceptBinding, BindableTable, _PrimitiveType, FilterBy, ReferentConceptBinding, _get_transform_types, \
    BindableColumn
from relationalai.early_access.dsl.codegen.helpers import reference_entity, construct_entity
from relationalai.early_access.dsl.orm.constraints import Unique
from relationalai.early_access.dsl.orm.relationships import Role
from relationalai.early_access.metamodel.util import OrderedSet


def _get_map_madlib_and_name(name_prefix: str, value_concept: qb.Concept) -> tuple[str, str]:
    """
    Generates a name for the map based on the value concept.
    """
    name = f'{name_prefix}_row_to_{value_concept}'
    madlib = f'{name} {{row_id:RowId}} {{val:{value_concept}}}'
    return madlib, name


class GeneratedRelation(qb.Relationship):
    """
    A class representing a relation generated from a model.
    """
    def __init__(self, madlib, model, name):
        super().__init__(madlib, model=model.qb_model(), short_name=name)

    def __repr__(self):
        return f'GeneratedRelation({self._name})'


class InternallyGeneratedRelation(GeneratedRelation):
    """
    A class representing a relation generated by analyzing the model.
    """
    def __init__(self, madlib, dsl_model, name):
        super().__init__(madlib, dsl_model, name)


def _filtering_view(row, filter_by: Optional[FilterBy], column_ref=None):
    """
    Generates filtering atoms for the map if the binding has any filters, which must be simple
    filters on the attribute view of the same table.
    """
    if not filter_by:
        return where()
    if isinstance(filter_by, _PrimitiveType):
        assert column_ref is not None, 'Attribute must be provided if filter_by is a PrimitiveFilterBy'
        atoms = [_primitive_filtering_atom(column_ref, filter_by)]
    elif isinstance(filter_by, qb.Expression):
        atoms = [_expr_filtering_atom(row, filter_by)]
    elif isinstance(filter_by, Iterable):
        atoms = [_expr_filtering_atom(row, filter_expr) for filter_expr in filter_by]
    elif isinstance(filter_by, Not):
        # trivial negation is supported, either on a single column or a trivial expression
        atoms = [_negation_atom(row, filter_by)]
    else:
        raise TypeError(f'Expected a PrimitiveFilterBy, Expression, or Iterable of Expressions, got {type(filter_by)}')
    return where(*atoms)

def _primitive_filtering_atom(column_ref, condition: _PrimitiveType):
    """
    Generates a filtering atom for the map based on a primitive filter.
    """
    if isinstance(condition, Enum):
        condition = condition.name
    return where(
        column_ref == condition
    )

def _expr_from_filter(row, condition: Union[BindableColumn, qb.Expression]):
    """
    Generates a filtering atom for the map based on a condition.
    """
    if isinstance(condition, qb.Expression):
        return _expr_filtering_atom(row, condition)
    elif isinstance(condition, BindableColumn):
        return _column_filtering_atom(row, condition)
    else:
        raise TypeError(f'Expected an Expression or a BindableColumn, got {type(condition)}')

def _expr_filtering_atom(row, condition: qb.Expression):
    params = condition._params
    if len(params) != 2:
        raise ValueError(f'Expected a condition with two parameters, got {len(params)}: {condition}')
    column, value = params
    if isinstance(value, Enum):
        value = value.name
    orig = column.type().ref()
    return where(
        column(row, orig),
        qb.Expression(condition._op, orig, value)
    )

def _column_filtering_atom(row, column: BindableColumn):
    """
    Defines an expression which could be used to express `exists(val: column(row, val))` for a given column.
    """
    orig = column.type().ref()
    return where(
        column(row, orig)
    )

def _negation_atom(row, condition: Not):
    """
    Generates a negation atom based on the filter: either a column or a trivial expression.
    """
    conditions = condition._args
    return where(
        not_(
            *[_expr_from_filter(row, condition) for condition in conditions]
        )
    )


class AbstractMap(GeneratedRelation):
    """
    A class representing an abstract map.
    """
    def __init__(self, madlib, dsl_model, name):
        super().__init__(madlib, dsl_model, name)

    def row_ref(self):
        assert len(self._field_refs) == 2, 'AbstractMap must have exactly two field references'
        row_ref = self._field_refs[0]
        assert row_ref._thing._name == 'RowId', 'First field must be of type RowId'
        return row_ref

    def value_player(self) -> qb.Concept:
        """
        Returns the value player of the map.
        This is the concept that the map maps to.
        """
        raise NotImplementedError()

    def formula(self, row_ref):
        rez_val_ref = self.value_player().ref()
        return rez_val_ref, [self(row_ref, rez_val_ref)]


class RoleMap(AbstractMap):
    """
    A class representing a base role map relation.
    """

    def __init__(self, madlib, model, name, role):
        super().__init__(madlib, model, name)
        self._role = role

    def role(self):
        """
        Returns the role of the role map.
        """
        return self._role

    def value_player(self):
        """
        Returns the value player of the role map.
        """
        return self._role.player()

    @abstractmethod
    def column(self) -> BindableAttribute:
        """
        Returns the bindable column associated with this role map.
        """
        raise NotImplementedError("Subclasses must implement this method")

    def table(self) -> BindableTable:
        """
        Returns the table associated with the role map.
        """
        return self.column().table

    def __repr__(self):
        return f'RoleMap({self._name})'


class InlineValueMap:
    """
    A non-QB relationship class that represents an inline value map.
    """
    def __init__(self, model, binding: Binding, role):
        self._binding = binding
        self._role = role
        self._value_player = role.player()
        self._initialize()

    def _initialize(self):
        role = self._role
        if not role.player()._is_primitive():
            raise TypeError(f'Cannot construct a value map for {role}, concept is not a value type')

        table_name = self._binding.column.table.physical_name()
        concept = role.player()
        _, self._name = _get_map_madlib_and_name(table_name, concept)
        self._has_transforms = bool(self._binding.transform_with)

    def __call__(self, *args, **kwargs):
        """
        Returns a QB expression that can be used in a query.
        """
        # assert that len(args) is 2, as we expect row and value variables
        if len(args) != 2:
            raise ValueError(f'Expected 2 arguments, got {len(args)}: {args}')

        row_ref, rez_val_ref = args
        _, body_atoms = self.construct_body_formula(row_ref, rez_val_ref)
        return where(*body_atoms)

    def construct_body_formula(self, row_ref, rez_val_ref):
        """
        Constructs the body formula for the value map.
        This is used to generate the head and the body of the relation.
        """
        col_val_ref = self.column().type().ref() if self._has_transforms else rez_val_ref
        filter_atoms = _filtering_view(row_ref, self._binding.filter_by, col_val_ref)
        return self._construct_body_atoms(row_ref, col_val_ref, rez_val_ref, filter_atoms)

    def _construct_body_atoms(self, row_ref, col_val_ref, rez_val_ref, filter_atoms):
        result_type = self.value_player()
        if self._has_transforms:
            transform_atoms = self._generate_transform_body(col_val_ref, rez_val_ref)
            return [row_ref, result_type(rez_val_ref)], [
                where(
                    self._binding.column(row_ref, col_val_ref),
                    transform_atoms,
                    filter_atoms
                )]
        else:
            # col_val_ref and rez_val_ref are the same handles here
            return [row_ref, result_type(col_val_ref)], [
                self._binding.column(row_ref, col_val_ref),
                filter_atoms,
            ]

    def _generate_transform_body(self, col_val_ref, rez_val_ref):
        transformations = self._binding.transform_with
        if transformations:
            transformations = (transformations,) if not isinstance(transformations, tuple) else transformations
            return self._chain_transformations(col_val_ref, rez_val_ref, transformations)
        raise ValueError("Incorrect state, cannot apply binding transformations with empty value converter(s)")

    def _chain_transformations(self, input_val_ref, output_val_ref, transformations):
        if len(transformations) == 1:
            return self._apply_transformations(input_val_ref, output_val_ref, transformations[0])
        intermediate_vars = [self._get_transform_output_type(tr) for tr in transformations[:-1]]
        all_vars = [input_val_ref, *intermediate_vars, output_val_ref]
        return where(
            *[self._apply_transformations(all_vars[idx], all_vars[idx+1], tr)
              for idx, tr in enumerate(transformations)]
        )

    @staticmethod
    def _get_transform_output_type(transformer):
        _, output_type = _get_transform_types(transformer)
        return output_type

    @staticmethod
    def _apply_transformations(input_var, output_var, transformer):
        if isinstance(transformer, qb.Relationship):
            return transformer(input_var, output_var)
        elif isinstance(transformer, Callable):
            return output_var == transformer(input_var)
        else:
            raise TypeError(f'Expected a Relationship or Callable, got {type(transformer)}')

    def value_player(self):
        """
        Returns the value player of the inline value map.
        """
        return self._value_player

    def column(self) -> BindableAttribute:
        """
        Returns the bindable column associated with this inline value map.
        """
        return self._binding.column

    def table(self) -> BindableTable:
        """
        Returns the table associated with the inline value map.
        """
        return self.column().table

    def row_ref(self):
        """
        Returns the row reference for the inline value map.
        """
        return self.column().row_ref()

    def binding(self) -> Binding:
        """
        Returns the binding of the value map.
        """
        return self._binding

    def formula(self, row_ref):
        rez_val_ref = self.value_player().ref()
        return rez_val_ref, [self(row_ref, rez_val_ref)]

    def __repr__(self):
        return f'@inline ValueMap({self._name})'


class ValueMap(RoleMap):
    """
    A class representing a value map relation.
    """

    def __init__(self, model, binding: Binding, role):
        # =
        # Skip QB relationship initialization if *inline* is True, as __call__ will return a QB expression
        # =
        madlib, name = self._handle_params(binding, role)
        super().__init__(madlib, model, name, role)
        self._binding = binding
        self._inline_value_map = InlineValueMap(model, binding, role)
        self._generate_body()

    @staticmethod
    def _handle_params(binding, role):
        if not role.player()._is_primitive():
            raise TypeError(f'Cannot construct a value map for {role}, concept is not a value type')

        table_name = binding.column.table.physical_name()
        concept = role.player()
        return _get_map_madlib_and_name(table_name, concept)

    def _generate_body(self):
        row_ref = self.row_ref()
        has_transforms = bool(self._binding.transform_with)
        rez_val_ref = self.value_player().ref() if has_transforms else self.column().type().ref()

        head_vars, body_atoms = self._inline_value_map.construct_body_formula(row_ref, rez_val_ref)

        define(self(*head_vars)).where(*body_atoms)

    def value_player(self):
        return self._inline_value_map.value_player()

    def column(self) -> BindableAttribute:
        return self._inline_value_map.column()

    def binding(self) -> Binding:
        return self._inline_value_map.binding()

    def __repr__(self):
        return f'ValueMap({self._name})'


class MaterializedEntityMap(AbstractMap):
    """
    A class representing a materialized entity map relation.
    This is used to materialize an entity map from a set of role maps.
    """

    def __init__(self, model, inline_entity_map: 'AbstractInlineEntityMap'):
        super().__init__(inline_entity_map.madlib(), model, inline_entity_map.short_name())
        self._inline_entity_map = inline_entity_map
        self._generate_body()

    def _generate_body(self):
        # materialize the population of the entity map as necessary
        self._inline_entity_map.materialize_population()

        row_ref = self.row_ref()
        rez_val_ref, body_formula = self._inline_entity_map.formula(row_ref)
        where(*body_formula).define(self(row_ref, rez_val_ref))

    def value_player(self):
        """
        Returns the result value player of the materialized entity map.
        This is the same as the value player of the inline entity map.
        """
        return self._inline_entity_map.value_player()

    def column(self) -> BindableAttribute:
        return self._inline_entity_map.column()

    def table(self) -> BindableTable:
        """
        Returns the table associated with the materialized entity map.
        """
        return self._inline_entity_map.table()

    def __repr__(self):
        # replace @inline with @materialized
        return self._inline_entity_map.__repr__().replace('@inline', '@materialized')


class AbstractInlineEntityMap(ABC):
    """
    A non-QB relationship class that represents an inline entity map.
    This is used to construct entities from a set of role maps.
    """

    def __init__(self, model, madlib, name):
        self._model = model
        self._madlib = madlib
        self._name = name

    def __call__(self, *args, **kwargs):
        if len(args) != 1:
            raise ValueError(f'Expected a single argument to a Concept Map, got {len(args)}: {args}')

        row_ref = args[0]
        return self._construct_body_formula(row_ref)

    def madlib(self):
        """
        Returns the madlib of the inline entity map.
        """
        return self._madlib

    def short_name(self):
        """
        Returns the name of the inline entity map.
        """
        return self._name

    @abstractmethod
    def value_player(self):
        """
        Returns the value player of the role map.
        """
        raise NotImplementedError("Subclasses must implement this method")

    @abstractmethod
    def column(self) -> BindableAttribute:
        """
        Returns the bindable column associated with this role map.
        """
        raise NotImplementedError("Subclasses must implement this method")

    def table(self) -> BindableTable:
        """
        Returns the table associated with the role map.
        """
        return self.column().table

    @abstractmethod
    def row_ref(self):
        """
        Returns the row reference for the inline entity map, should be typically a row reference of an arbitrary
        inline entity or value map that is used to construct the entity, all the way to a column.
        """
        raise NotImplementedError("Subclasses must implement this method")

    def formula(self, row_ref):
        return self._construct_body_formula(row_ref)

    @abstractmethod
    def _construct_body_formula(self, row_ref) -> tuple[Union[ConceptConstruct, Ref, RelationshipRef], list[typing.Any]]:
        """
        Constructs the body formula for the inline entity map.
        This is used to generate the body of the relation.
        """
        pass

    def materialize(self):
        return MaterializedEntityMap(self._model, self)

    @abstractmethod
    def materialize_population(self):
        """
        Materializes the population of the entity map, using the body of the entity map.
        This has to check the conditions and only update the population if those are met.
        """
        pass


class SimpleConstructorEntityMap(AbstractInlineEntityMap):
    """
    A class representing an entity map relation.
    """

    def __init__(self, model, binding: Binding, role: Role, identifier_to_role_map: OrderedDict[Unique, 'RoleMap']):
        madlib, name = self._handle_params(binding, role)
        super().__init__(model, madlib, name)
        self._binding = binding
        self._role = role
        self._identifier_to_role_map = identifier_to_role_map
        self._reference_role_map = list(identifier_to_role_map.items())[-1][1]  # last role map is the reference one

    @staticmethod
    def _handle_params(binding, role):
        if role.player()._is_primitive():
            raise TypeError(f'Cannot construct an entity map for {role}, concept is not an entity type')

        table_name = binding.column.table.physical_name()
        concept = role.player()
        return _get_map_madlib_and_name(table_name, concept)

    def _construct_body_formula(self, row_ref):
        concept = self._role.player()
        # and populate the entity map, role_map is always the last one in the identifier_to_role_map
        value = self._reference_role_map.value_player().ref()
        body_atoms = [
            self._reference_role_map(row_ref, value),
            entity_ref := reference_entity(concept, value),
            self._concept_population_atom(entity_ref)
        ]
        return entity_ref, body_atoms

    def _concept_population_atom(self, entity):
        """
        Generates the *optional* population atom for the entity map.
        Only used for bindings that have a FK.
        """
        if self._should_reference():
            return where(
                self._role.player()(entity)
            )
        else:
            return where()

    def _should_reference(self) -> bool:
        return self._reference_role_map.column().references_column is not None

    def value_player(self):
        """
        Returns the value player of the entity map.
        """
        return self._role.player()

    def column(self) -> BindableAttribute:
        """
        Returns the bindable column associated with this entity map.
        """
        return self._binding.column

    def row_ref(self):
        return self._reference_role_map.row_ref()

    def materialize_population(self):
        row_ref = self.row_ref()
        concept = self._role.player()
        if not self._should_reference():
            # then create entities
            role_maps = self._identifier_to_role_map.values()
            map_value_refs = [role_map.value_player().ref() for role_map in role_maps]
            where(
                *[role_map(row_ref, value) for role_map, value in zip(role_maps, map_value_refs)]
            ).define(construct_entity(concept, *map_value_refs))

    def __repr__(self):
        return f'@inline CtorEntityMap({self._name})'


class ReferentEntityMap(AbstractInlineEntityMap):
    """
    A class representing a referent entity map relation.
    """

    def __init__(self, model, binding: Binding, role: Role, constructing_role_map: 'Map'):
        madlib, name = self._handle_params(binding, role)
        super().__init__(model, madlib, name)
        self._binding = binding
        self._role = role
        self._constructing_role_map = constructing_role_map

    @staticmethod
    def _handle_params(binding, role):
        if role.player()._is_primitive():
            raise TypeError(f'Cannot construct an entity map for {role}, concept is not an entity type')

        table_name = binding.column.table.physical_name()
        concept = role.player()
        return _get_map_madlib_and_name(f'{table_name}_ref', concept)

    def _construct_body_formula(self, row_ref):
        concept = self._role.player()
        orig_entity_ref, subformula_atoms = self._constructing_role_map.formula(row_ref)

        entity_ref = reference_entity(concept, orig_entity_ref)
        return entity_ref, [
            *subformula_atoms,
            entity_ref,
            concept(entity_ref)
        ]

    def value_player(self):
        return self._role.player()

    def binding(self) -> Binding:
        """
        Returns the binding of the entity map.
        """
        return self._binding

    def column(self) -> BindableAttribute:
        """
        Returns the bindable column associated with this entity map.
        """
        return self._binding.column

    def row_ref(self):
        return self._constructing_role_map.row_ref()

    def materialize_population(self):
        """
        Referent entity maps do not need to materialize population, as they always reference existing entities.
        """
        pass

    def __repr__(self):
        return f'@inline ReferentEntityMap({self._name})'


class CompositeEntityMap(AbstractInlineEntityMap):
    """
    A class representing a composite entity map relation.

    Takes a sequence of EntityMaps and constructs a composite entity type.
    """

    def __init__(self, model, entity_concept: qb.Concept, *role_maps: 'Map'):
        madlib, name = self._handle_params(entity_concept, *role_maps)
        super().__init__(model, madlib, name)
        self._entity_concept = entity_concept
        self._role_maps = list(role_maps)

    def _handle_params(self, entity_concept: qb.Concept, *role_maps: 'Map'):
        if entity_concept._is_primitive():
            raise TypeError(f'Cannot construct a composite entity map for {entity_concept},'
                            f' concept is not an entity type')
        if len(role_maps) < 2:
            raise ValueError('CompositeEntityMap requires at least two EntityMaps')

        role_map = role_maps[0]
        table = role_map.table()
        self._table = table
        return _get_map_madlib_and_name(table.physical_name(), entity_concept)

    def _construct_body_formula(self, row_ref):
        var_refs, body_atoms = self._body_formula(row_ref)

        # the rule to populate the entity map
        entity_ref = reference_entity(self._entity_concept, *var_refs)
        body_atoms.append(entity_ref)  # TODO should this be added?
        return entity_ref, body_atoms

    @staticmethod
    def _should_reference() -> bool:
        """
        Determines if the composite entity map should reference a column.
        Currently not supported fully.
        """
        return False

    def _body_formula(self, row_ref):
        var_refs, subformulas = zip(*(role_map.formula(row_ref) for role_map in self._role_maps))
        return var_refs, [
            *(atom for subformula in subformulas for atom in subformula)
        ]

    def value_player(self):
        """
        Returns the value player of the composite entity map.
        """
        return self._entity_concept

    def table(self) -> BindableTable:
        """
        Returns the table associated with the composite entity map.
        """
        return self._table

    def column(self) -> BindableAttribute:
        raise ValueError('CompositeEntityMap does not index a single column')

    def row_ref(self):
        # take arbitrary row reference from the first role map
        role_map = self._role_maps[0]
        return role_map.row_ref()

    def materialize_population(self):
        row_ref = self.row_ref()
        var_refs, body_atoms = self._body_formula(row_ref)
        if not self._should_reference():
            # construct entities
            where(
                *body_atoms,
                entity_ref := construct_entity(self._entity_concept, *var_refs),
            ).define(entity_ref)

    def __repr__(self):
        return f'@inline CompositeEntityMap({self._name})'


AbstractEntityMap = Union['SimpleConstructorEntityMap', 'ReferentEntityMap', 'CompositeEntityMap', 'UnionEntityMap']
Map = Union['ValueMap', 'InlineValueMap', 'UnionEntityMap', 'AbstractInlineEntityMap', 'MaterializedEntityMap']

class EntitySubtypeMap(AbstractInlineEntityMap):
    """
    A class representing an entity subtype map relation.
    """

    def __init__(self, model, binding: Union[IdentifierConceptBinding, SubtypeConceptBinding, ReferentConceptBinding],
                 ctor_entity_map: Union[AbstractInlineEntityMap, MaterializedEntityMap]):
        # type of the entity subtype map is the parent type coming from the ctor_entity_map
        madlib, name = self._handle_params(binding, ctor_entity_map.value_player())
        super().__init__(model, madlib, name)
        self._binding = binding
        self._ctor_entity_map = ctor_entity_map

    @staticmethod
    def _handle_params(binding: Union[IdentifierConceptBinding, SubtypeConceptBinding, ReferentConceptBinding],
                       result_concept: qb.Concept):
        table = binding.column.table
        entity_concept = binding.entity_type
        # check that the entity concept is a subtype of the result concept
        assert entity_concept._isa(result_concept), f"Entity concept {entity_concept} must be a subtype of result concept {result_concept}"
        name = f'{table.physical_name()}_row_to_{entity_concept}'
        madlib = f'{name} {{row_id:RowId}} {{val:{result_concept}}}'
        return madlib, name

    def _construct_body_formula(self, row_ref):
        parent_type_ref, body_atoms = self._generate_body_atoms(row_ref)

        # note: have to return the parent type here, as otherwise it gets a join on the subtype population
        #       but doesn't cast to the type of the subtype...
        return parent_type_ref, body_atoms

    def _generate_body_atoms(self, row_ref):
        filtering_atom = _filtering_view(row_ref, self._binding.filter_by)
        parent_type_ref, subformula_atoms = self._ctor_entity_map.formula(row_ref)
        return parent_type_ref, [
            *subformula_atoms,
            filtering_atom
        ]

    def binding(self):
        """
        Returns the binding of the entity subtype map.
        """
        return self._binding

    def value_player(self):
        """
        Returns the subtype of the entity subtype map.
        """
        return self._binding.entity_type

    def column(self) -> BindableAttribute:
        """
        Returns the bindable column associated with this entity subtype map.
        """
        return self._binding.column

    def table(self) -> BindableTable:
        """
        Returns the table associated with the entity subtype map.
        """
        return self._binding.column.table

    def row_ref(self):
        return self._ctor_entity_map.row_ref()

    def materialize_population(self):
        row_ref = self.row_ref()
        subtype = self._binding.entity_type
        parent_type_ref, body_atoms = self._generate_body_atoms(row_ref)
        # TODO instead of Person.new maybe try Employee.new to construct these?
        # derive subtype population
        where(*body_atoms).define(subtype(parent_type_ref))

    def __repr__(self):
        return f'@inline EntitySubtypeMap({self._name})'


class UnionEntityMap(AbstractMap):
    """
    A class representing a union entity map relation.
    """

    def __init__(self, model, entity_concept: qb.Concept, *entity_maps: AbstractEntityMap, generate_population: bool=False):
        madlib, name = self._handle_params(entity_concept, *entity_maps)
        super().__init__(madlib, model, name)
        self._entity_type = entity_concept
        self._entity_maps = OrderedSet().update(entity_maps)
        self._generate_population = generate_population
        self._generate_body()

    def _handle_params(self, entity_concept: qb.Concept, *entity_maps: AbstractEntityMap):
        if len(entity_maps) == 0:
            raise ValueError('UnionEntityMap requires at least one EntityMap')
        # pick an arbitrary entity map to get the table, as they must all have the same
        table = entity_maps[0].table()
        self._table = table
        return _get_map_madlib_and_name(table.physical_name(), entity_concept)

    def _generate_body(self):
        for entity_map in self._entity_maps:
            self._generate_body_rule(entity_map)

    def _generate_body_rule(self, entity_map: AbstractEntityMap):
        row_ref = entity_map.row_ref()
        rez_val_ref, subformula_atoms = entity_map.formula(row_ref)

        # derive union entity map
        where(
            *subformula_atoms
        ).define(self(row_ref, rez_val_ref))

    def value_player(self):
        """
        Returns the type of the entity map.
        """
        return self._entity_type

    def table(self):
        """
        Returns the table associated with the entity map.
        """
        return self._table

    def update(self, entity_map: AbstractEntityMap):
        """
        Updates the union entity map with a new entity map.
        """
        if entity_map in self._entity_maps:
            return
        self._entity_maps.add(entity_map)
        self._generate_body_rule(entity_map)

    def materialize_population(self):
        if self._generate_population:
            # derive type population
            for entity_map in self._entity_maps:
                row_ref = entity_map.row_ref()
                rez_val_ref, subformula_atoms = entity_map.formula(row_ref)
                where(
                    *subformula_atoms
                ).define(self._entity_type(rez_val_ref))

    def __repr__(self):
        return f'UnionEntityMap({self._name})'



# Find shortest paths and walks in a graph

from relationalai.early_access.builder import std
from relationalai.early_access.builder import Integer
from relationalai.early_access.builder import where, define, rank, sum, not_
from relationalai.early_access.builder.std.integers import int64



from ..graph import Graph
from ..utilities.prefix_sum import linear_prefix_sum_with_groupby
from .usp import compute_usp, compute_nsp_from_usp, compute_uw, compute_nw_from_uw



def range_int64(start, stop, step=1):
    """Helper for std.range with int64 arguments."""
    return std.range(int64(start), int64(stop), int64(step))


def find_shortest_paths(g: Graph, Source, Target, max_length=None, num_paths=None):
    Node = g.Node

    u, v, src = Node.ref(), Node.ref(), Node.ref()
    i, val, lo, hi = Integer.ref(), Integer.ref(), Integer.ref(), Integer.ref()
    level, n, m, path_num = Integer.ref(), Integer.ref(), Integer.ref(), Integer.ref()
    d, r, s, tot = Integer.ref(), Integer.ref(), Integer.ref(), Integer.ref()

    ord_source = g.model.Relationship(f"ord_source {{Integer}}th source node {{{Node}}}")
    total_paths = g.model.Relationship("total_paths {Integer}")
    int_source = g.model.Relationship(f"int_source paths from {{{Node}}} numbered {{Integer}} to {{Integer}}")
    paths_listing = g.model.Relationship(f"paths_listing {{{Node}}} {{Integer}} {{Integer}} {{Integer}}")
    paths = g.model.Relationship(f"paths {{{Integer}}} has {{{Integer}}}th node {{{Node}}}")
    neighbor = g.model.Relationship(f"neighbor {{{Node}}} neighbor {{Integer}} is {{{Node}}}")
    edge_nsp = g.model.Relationship(f"edge_nsp num shortest from {{{Node}}} through {{Integer}} neighbor {{Integer}}")
    acc_nsp = g.model.Relationship(f"acc_nsp num shortest from {{{Node}}} neighbor from 1 to {{Integer}} is {{Integer}}")
    nsp_interval = g.model.Relationship(f"nsp_interval {{{Node}}} {{Integer}} {{Integer}} {{Integer}}")
    path_routing = g.model.Relationship(f"path_routing {{{Node}}} {{Integer}} {{{Node}}} {{Integer}}")

    usp, Boundary = compute_usp(g, Source, Target, max_length)

    where(usp(u, v), r == rank(v).per(u)).define(neighbor(u, r, v))

    nsp = compute_nsp_from_usp(g, usp, Source, Target, Boundary)

    define(edge_nsp(u, i, val)).where(
        neighbor(u, i, v),
        nsp(v, val)
    )

    acc_nsp = linear_prefix_sum_with_groupby(edge_nsp)

    define(nsp_interval(u, i, lo, hi)).where(
        i == 1,
        lo == 1,
        acc_nsp(u, i, hi)
    )
    define(nsp_interval(u, i, lo, hi)).where(
        acc_nsp(u, i - 1, lo - 1),
        acc_nsp(u, i, hi)
    )

    define(ord_source(r, u)).where(
        Source(u),
        usp(u, v),
        r == rank(u)
    )

    define(ord_source(r, u)).where(
        Source(u),
        Target(u),
        r == rank(u)
    )

    if num_paths is not None:
        define(total_paths(num_paths))
    else:
        define(total_paths(sum(src, n))).where(nsp(src, n))

    define(int_source(src, n, m)).where(
        total_paths(tot),
        ord_source(1, src),
        n == 1,
        nsp(src, level),
        n <= tot,
        tot <= level,
        m == tot
    )

    define(int_source(src, n, m)).where(
        total_paths(tot),
        ord_source(1, src),
        n == 1,
        nsp(src, level),
        level < tot,
        m == level
    )

    define(int_source(src, n, m)).where(
        total_paths(tot),
        ord_source(d, src),
        ord_source(d-1, u),
        int_source(u, r, s),
        n == s + 1,
        nsp(src, level),
        n <= tot,
        tot <= s + level,
        m == tot
    )

    define(int_source(src, n, m)).where(
        total_paths(tot),
        ord_source(d, src),
        ord_source(d-1, u),
        int_source(u, r, s),
        n == s + 1,
        nsp(src, level),
        s + level < tot,
        m == s + level
    )

    where(
        int_source(src, n, m),
        path_num == range_int64(n, m + 1, 1),
        r == path_num - n + 1,
    ).define(paths_listing(src, r, 0, path_num))

    where(
        paths_listing(u, m, level - 1, path_num),
        path_routing(u, m, v, n)
    ).define(paths_listing(v, n, level, path_num))

    where(
        nsp_interval(u, i, lo, hi),
        m == range_int64(lo, hi + 1, 1),
        neighbor(u, i, v),
        n == m - lo + 1
    ).define(path_routing(u, m, v, n))

    define(paths(path_num, level, v)).where(
        paths_listing(v, n, level, path_num)
    )

    return paths


def find_walks(g: Graph, Source, Target, max_length, num_paths=None):
    Node = g.Node

    u, v, src = Node.ref(), Node.ref(), Node.ref()
    i, val, lo, hi = Integer.ref(), Integer.ref(), Integer.ref(), Integer.ref()
    level, n, m, path_num = Integer.ref(), Integer.ref(), Integer.ref(), Integer.ref()
    d, r, s, tot = Integer.ref(), Integer.ref(), Integer.ref(), Integer.ref()

    ord_source = g.model.Relationship(f"ord_source {{Integer}}th source node {{{Node}}}")
    total_paths = g.model.Relationship("total_paths {Integer}")
    int_source = g.model.Relationship(f"int_source paths from {{{Node}}} numbered {{Integer}} to {{Integer}}")
    paths_listing = g.model.Relationship(f"paths_listing {{{Node}}} {{Integer}} {{Integer}} {{Integer}}")
    paths = g.model.Relationship(f"paths {{{Integer}}} has {{{Integer}}}th node {{{Node}}}")
    neighbor = g.model.Relationship(f"neighbor pair {{{Node}}} {{Integer}} neighbor {{Integer}} is {{{Node}}}")
    edge_nsp = g.model.Relationship(f"edge_nsp num shortest from pair {{{Node}}} {{Integer}} through {{Integer}} neighbor {{Integer}}")
    acc_nsp = g.model.Relationship(f"acc_nsp num shortest from {{{Node}}} neighbor from 1 to {{Integer}} is {{Integer}}")
    nsp_interval = g.model.Relationship(f"nsp_interval {{{Node}}} {{Integer}} {{Integer}} {{Integer}} {{Integer}}")
    path_routing = g.model.Relationship(f"path_routing {{{Node}}} {{Integer}} {{Integer}} {{{Node}}} {{Integer}}")

    uw, Boundary = compute_uw(g, Source, Target, max_length)


    where(uw(u, n, v), r == rank(v).per(u, n)).define(neighbor(u, n, r, v))

    nw = compute_nw_from_uw(g, uw, Target, Boundary)

    define(edge_nsp(u, n, i, val)).where(
        neighbor(u, n, i, v),
        nw(v, n + 1, val)
    )

    acc_nsp = linear_prefix_sum_with_groupby(edge_nsp)

    define(nsp_interval(u, n, i, lo, hi)).where(
        i == 1,
        lo == 1,
        acc_nsp(u, n, i, hi)
    )
    define(nsp_interval(u, n, i, lo, hi)).where(
        acc_nsp(u, n, i - 1, lo - 1),
        acc_nsp(u, n, i, hi)
    )

    define(ord_source(r, u)).where(
        Source(u),
        uw(u, 0, v),
        r == rank(u)
    )

    define(ord_source(r, u)).where(
        Source(u),
        Target(u),
        r == rank(u)
    )

    if num_paths is not None:
        define(total_paths(num_paths))
    else:
        define(total_paths(sum(src, 0, n))).where(nw(src, 0, n))

    define(int_source(src, n, m)).where(
        total_paths(tot),
        ord_source(1, src),
        n == 1,
        nw(src, 0, level),
        n <= tot,
        tot <= level,
        m == tot
    )

    define(int_source(src, n, m)).where(
        total_paths(tot),
        ord_source(1, src),
        n == 1,
        nw(src, 0, level),
        level < tot,
        m == level
    )

    define(int_source(src, n, m)).where(
        total_paths(tot),
        ord_source(d, src),
        ord_source(d-1, u),
        int_source(u, r, s),
        n == s + 1,
        nw(src, 0, level),
        n <= tot,
        tot <= s + level,
        m == tot
    )

    define(int_source(src, n, m)).where(
        total_paths(tot),
        ord_source(d, src),
        ord_source(d-1, u),
        int_source(u, r, s),
        n == s + 1,
        nw(src, 0, level),
        s + level < tot,
        m == s + level
    )

    where(
        int_source(src, n, m),
        path_num == range_int64(n, m + 1, 1),
        r == path_num - n + 1,
    ).define(paths_listing(src, r, 0, path_num))

    where(
        paths_listing(u, m, level - 1, path_num),
        path_routing(u, level - 1, m, v, n),
        not_(Target(u))
    ).define(paths_listing(v, n, level, path_num))

    where(
        paths_listing(u, m, level - 1, path_num),
        path_routing(u, level - 1, m - 1, v, n),
        Target(u),
        m >= 2
    ).define(paths_listing(v, n, level, path_num))

    where(
        nsp_interval(u, level, i, lo, hi),
        m == range_int64(lo, hi + 1, 1),
        neighbor(u, level, i, v),
        n == m - lo + 1
    ).define(path_routing(u, level, m, v, n))

    define(paths(path_num, level, v)).where(
        paths_listing(v, n, level, path_num)
    )

    return paths

# Explicitly export relevant classes and functions
__all__ = []



from __future__ import annotations
from functools import lru_cache
import os
import sys
import warnings

from relationalai.environments.base import SourcePos

from .base import RuntimeEnvironment, SourceInfo, find_external_frame, handle_rai_exception, handle_rai_warning, patch

cwd = os.getcwd()

# relpath is a shockingly slow operation, so caching it
@lru_cache(maxsize=128)
def relative_path(path: str) -> str:
    try:
        return os.path.relpath(path, cwd)
    except ValueError:
        return path

@lru_cache(maxsize=128)
def get_source_code(filename: str) -> str|None:
    """Get the source code of a file."""
    try:
        with open(filename, "r") as f:
            return f.read()
    except Exception:
        return None

class GenericEnvironment(RuntimeEnvironment):
    @classmethod
    def detect(cls) -> bool:
        return True

    def _to_source(self):
        caller_frame = find_external_frame()
        if not caller_frame:
            return

        caller_filename = caller_frame.f_code.co_filename
        caller_line = caller_frame.f_lineno
        relative_filename = relative_path(caller_filename)

        source_code = get_source_code(caller_filename)

        return (relative_filename, caller_line, source_code)

    def get_source(self, steps:int|None = None):
        source = self._to_source()
        if not source:
            return None
        (relative_filename, caller_line, source_code) = source
        return SourceInfo.from_source(relative_filename, caller_line, source_code)

    def get_source_pos(self, steps: int | None = None) -> SourcePos | None:
        source = self._to_source()
        if not source:
            return None
        (relative_filename, caller_line, source_code) = source
        return SourcePos(relative_filename, caller_line, source_code)

    def _patch(self):
        @patch(warnings, "showwarning")
        def _(original, message, category, filename, lineno, file=None, line=None):
            if not handle_rai_warning(message):
                original(message, category, filename, lineno, file, line)

        @patch(sys, "excepthook")
        def _(original, exc_type, exc_value, exc_traceback, quiet=False):
            handle_rai_exception(exc_value)
            original(exc_type, exc_value, exc_traceback)



from __future__ import annotations

from typing import Any
from relationalai.early_access.builder import builder as b, std

def make_name(*args, sep: str | None = "_"):
    if not args:
        raise ValueError("No arguments provided to `make_name`")
    elif len(args) == 1:
        arg = args[0]
        if isinstance(arg, str):
            return arg
        elif isinstance(arg, list):
            return make_name(*arg, sep=sep)
        else:
            return std.strings.string(arg)
    elif sep:
        str_args = []
        for a in args:
            str_args.append(std.strings.string(a))
            str_args.append(sep)
        str_args.pop()
    else:
        str_args = map(std.strings.string, args)
    return std.strings.concat(*str_args)

# TODO move to std? need to support normal logical evaluation.
def all_different(*args: Any) -> b.Aggregate:
    return b.Aggregate(b.Relationship.builtins["all_different"], *args)

def implies(left, right) -> b.Expression:
    return b.Expression(b.Relationship.builtins["implies"], left, right)



# Copyright 2024 RelationalAI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This module defines generic expressions (Exprs). These are
# extended for use in Schema and Type constraints in other
# modules.
#
from abc import abstractmethod

from relationalai.early_access.dsl.core.context import ContextStack

contextStack = ContextStack()


class Expr:

    # We can assert equality (==) between any pair of Exprs, though the
    # meaning varies depending on whether self is an ScalarExpr or
    # a RelationExpr
    #
    def __eq__(self, other): # type: ignore
        contextStack.root_context().build_comparison(self, "=", other)

    # Returns a printable form of this Expr.
    #
    @abstractmethod
    def display(self) -> str: pass

    # Returns a copy of this Expr with renaming applied to replace
    # any variable that occurs in the renaming dictionary, which is
    # keyed by the names of the variables to rename. For instance,
    # if expr had the form (x + 2), then:
    #
    #    expr.rename({"x": y})
    #
    # would have the form (y + 2)
    #
    @abstractmethod
    def rename(self, renaming) -> 'Expr': pass

    # Returns a copy of this Expr with vmap applied to replace
    # any variable that occurs in the vmap dictionary, which is
    # keyed by the identifier of the variables to revar. For
    # instance, if expr had the form (x + 2) and the entityid
    # of x was 100, then:
    #
    #    expr.revar({100: y})
    #
    # would have the form (y + 2)
    #
    @abstractmethod
    def revar(self, vmap) -> 'Expr': pass

    # Returns a dictionary of ScalarVariables referenced by this Expr,
    # with variable names (strings) as keys
    #
    def scalar_refs(self): return {}

    # Returns an integer (hash) that identifies this Expr, allowing us
    # to treat Exprs sort of like entity types for which we had declared
    # a preferred identifier so as to allow different physical objects
    # to unify to the same entity. Contrast this with the "id" method
    # which would not allow us to do this unification.
    #
    @abstractmethod
    def entityid(self) -> int: pass

    # This predicate returns True if this Expr refers to a variable
    # with this varname and False otherwise.
    #
    @abstractmethod
    def refersto(self, varname: str) -> bool: pass

    # This predicate returns True if this Expr references any kind of
    # Relation and False otherwise.
    #
    def relational(self) -> bool: return False

    # This predicate returns True if this Expr is a (scalar or relation)
    # variable and False otherwise.
    #
    def variable(self) -> bool: return False


class Wildcard(Expr):
    def __init__(self): pass

    def display(self): return "_"

    def grounded_using(self, groundings): return True

    def refersto(self, varname: str): return False

    def rename(self, renaming): return self

    def revar(self, vmap): return self

    def substitute(self, bindings): return self

    def entityid(self): return hash(Wildcard)


# Declare this variable for use in Datalog-style rules
#
_ = Wildcard()



from relationalai.early_access.builder import Model, Integer, select, String, define

from relationalai.early_access.paths.graph import Graph
from relationalai.early_access.paths.api import path, star, match

# Create an automaton that matches paths with pattern: A(A*)B:
pattern = path("A", star("A"), "B")
edge_list = [(1, 2, "A"), (2, 3, "A"), (3, 4, "A"), (4, 5, "B"), (1, 5, "A")]

sources = [1]
targets = [5]

model = Model("my_paths")

# create the graph:
graph = Graph.from_edge_list(model, edge_list)
Node = graph.Node

# create source and target sets:
Source = model.Relationship("source3 {Node}")
Target = model.Relationship("target3 {Node}")

for src in sources:
    define(Source(graph.Node.new(id=src)))

for tgt in targets:
    define(Target(graph.Node.new(id=tgt)))

paths = match(graph, pattern, Source, Target)

i, k = Integer.ref(), Integer.ref()
graph_node = graph.Node.ref()
label = String.ref()

result = select(i, k, label, graph_node.id).where(
    paths(i, k, label, graph_node)
)

result.inspect()

# print(result)



from __future__ import annotations
from collections import defaultdict
import atexit
import re

from pandas import DataFrame
from typing import Any, Optional
import relationalai as rai

from relationalai import debugging
from relationalai.early_access.lqp import result_helpers
from relationalai.early_access.metamodel import ir, factory as f, executor as e
from relationalai.early_access.lqp.compiler import Compiler
from relationalai.early_access.lqp.types import lqp_type_to_sql
from lqp import print as lqp_print, ir as lqp_ir
from lqp.parser import construct_configure
from relationalai.early_access.lqp.ir import convert_transaction, validate_lqp
from relationalai.clients.config import Config
from relationalai.clients.snowflake import APP_NAME
from relationalai.clients.types import TransactionAsyncResponse
from relationalai.clients.util import IdentityParser
from relationalai.tools.constants import USE_DIRECT_ACCESS


class LQPExecutor(e.Executor):
    """Executes LQP using the RAI client."""

    def __init__(
        self,
        database: str,
        dry_run: bool = False,
        keep_model: bool = True,
        wide_outputs: bool = False,
        config: Config | None = None,
    ) -> None:
        super().__init__()
        self.database = database
        self.dry_run = dry_run
        self.keep_model = keep_model
        self.wide_outputs = wide_outputs
        self.compiler = Compiler()
        self.config = config or Config()
        self._resources = None
        self._last_model = None
        self._last_sources_version = (-1, None)

    @property
    def resources(self):
        if not self._resources:
            with debugging.span("create_session"):
                self.dry_run |= bool(self.config.get("compiler.dry_run", False))
                resource_class = rai.clients.snowflake.Resources
                if self.config.get("use_direct_access", USE_DIRECT_ACCESS):
                    resource_class = rai.clients.snowflake.DirectAccessResources
                self._resources = resource_class(dry_run=self.dry_run, config=self.config, generation=rai.Generation.QB)
                if not self.dry_run:
                    self.engine = self._resources.get_default_engine_name()
                    if not self.keep_model:
                        atexit.register(self._resources.delete_graph, self.database, True)
        return self._resources

    # Checks the graph index and updates it if necessary
    def prepare_data(self):
        # Has to happen first, so self.dry_run is populated.
        resources = self.resources

        if self.dry_run:
            return

        from relationalai.early_access.builder.snowflake import Table
        table_sources = Table._used_sources
        if not table_sources.has_changed(self._last_sources_version):
            return

        model = self.database
        app_name = resources.get_app_name()
        engine_name = self.engine
        engine_size = self.resources.config.get_default_engine_size()

        program_span_id = debugging.get_program_span_id()
        sources = [t._fqn for t in Table._used_sources]
        self._last_sources_version = Table._used_sources.version()

        assert self.engine is not None

        with debugging.span("poll_use_index", sources=sources, model=model, engine=engine_name):
            resources.poll_use_index(app_name, sources, model, self.engine, engine_size, program_span_id)

    def report_errors(self, problems: list[dict[str, Any]], abort_on_error=True):
        from relationalai import errors
        all_errors = []
        undefineds = []
        pyrel_errors = defaultdict(list)
        pyrel_warnings = defaultdict(list)

        for problem in problems:
            message = problem.get("message", "")
            report = problem.get("report", "")
            # TODO: we need to build source maps
            # path = problem.get("path", "")
            # source_task = self._install_batch.line_to_task(path, problem["start_line"]) or task
            # source = debugging.get_source(source_task) or debugging.SourceInfo()
            source = debugging.SourceInfo()
            severity = problem.get("severity", "warning")
            code = problem.get("code")

            if severity in ["error", "exception"]:
                if code == "UNDEFINED_IDENTIFIER":
                    match = re.search(r'`(.+?)` is undefined', message)
                    if match:
                        undefineds.append((match.group(1), source))
                    else:
                        all_errors.append(errors.RelQueryError(problem, source))
                elif "overflowed" in report:
                    all_errors.append(errors.NumericOverflow(problem, source))
                elif code == "PYREL_ERROR":
                    pyrel_errors[problem["props"]["pyrel_id"]].append(problem)
                elif abort_on_error:
                    all_errors.append(errors.RelQueryError(problem, source))
            else:
                if code == "ARITY_MISMATCH":
                    errors.ArityMismatch(problem, source)
                elif code == "IC_VIOLATION":
                    all_errors.append(errors.IntegrityConstraintViolation(problem, source))
                elif code == "PYREL_ERROR":
                    pyrel_warnings[problem["props"]["pyrel_id"]].append(problem)
                else:
                    errors.RelQueryWarning(problem, source)

        if abort_on_error and len(undefineds):
            all_errors.append(errors.UninitializedPropertyException(undefineds))

        if abort_on_error:
            for pyrel_id, pyrel_problems in pyrel_errors.items():
                all_errors.append(errors.ModelError(pyrel_problems))

        for pyrel_id, pyrel_problems in pyrel_warnings.items():
            errors.ModelWarning(pyrel_problems)


        if len(all_errors) == 1:
            raise all_errors[0]
        elif len(all_errors) > 1:
            raise errors.RAIExceptionSet(all_errors)

    def _export(self, txn_id: str, export_info: tuple, dest_fqn: str, actual_cols: list[str], declared_cols: list[str], update:bool):
        # At this point of the export, we assume that a CSV file has already been written
        # to the Snowflake Native App stage area. Thus, the purpose of this method is to
        # copy the data from the CSV file to the destination table.
        _exec = self.resources._exec
        dest_database, dest_schema, dest_table, _ = IdentityParser(dest_fqn, require_all_parts=True).to_list()
        filename = export_info[0]
        result_table_name = filename + "_table"

        with debugging.span("export", txn_id=txn_id, export_info=export_info, dest_table=dest_table):
            with debugging.span("export_to_result_schema"):
                # First, we need to persist from the CSV file to the results schema by calling the
                # `persist_from_stage` stored procedure. This step also cleans up the CSV file in
                # the stage area.
                column_fields = []
                for (col_name, col_type) in export_info[1]:
                    column_fields.append([col_name, lqp_type_to_sql(col_type)])

                # NOTE: the `str(column_fields)` depends on python formatting which surrounds
                # strings with single quotes. If this changes, or if we ever get a single quote in
                # the actual string, then we need to do something more sophisticated.
                exec_str = f"call {APP_NAME}.api.persist_from_stage('{txn_id}', '{filename}', '{result_table_name}', {str(column_fields)})"
                _exec(exec_str)

            with debugging.span("write_table"):
                # The result of the first step above is a table in the results schema,
                # {app_name}.results.{result_table_name}.
                # Second, we need to copy the data from the results schema to the
                # destination table. This step also cleans up the result table.
                out_sample = _exec(f"select * from {APP_NAME}.results.{result_table_name} limit 1;")
                if out_sample:
                    keys = set([k.lower() for k in out_sample[0].as_dict().keys()])
                    fields = []
                    ix = 0
                    for name in declared_cols:
                        if name in actual_cols:
                            field = f"col{ix:03} as \"{name}\"" if f"col{ix:03}" in keys else f"NULL as {name}"
                            ix += 1
                        else:
                            field = f"NULL as \"{name}\""
                        fields.append(field)
                    names = ", ".join(fields)
                    if not update:
                        _exec(f"""
                            BEGIN
                                -- Check if table exists
                                IF (EXISTS (
                                    SELECT 1
                                    FROM {dest_database}.INFORMATION_SCHEMA.TABLES
                                    WHERE table_schema = '{dest_schema}'
                                    AND table_name = '{dest_table}'
                                )) THEN
                                    -- Insert into existing table
                                    EXECUTE IMMEDIATE '
                                        BEGIN
                                            TRUNCATE TABLE {dest_fqn};
                                            INSERT INTO {dest_fqn}
                                            SELECT {names}
                                            FROM {APP_NAME}.results.{result_table_name};
                                        END;
                                    ';
                                ELSE
                                    -- Create table based on the SELECT
                                    EXECUTE IMMEDIATE '
                                        CREATE TABLE {dest_fqn} AS
                                        SELECT {names}
                                        FROM {APP_NAME}.results.{result_table_name};
                                    ';
                                END IF;
                                CALL {APP_NAME}.api.drop_result_table('{result_table_name}');
                            END;
                        """)
                    else:
                        _exec(f"""
                            begin
                                INSERT INTO {dest_fqn} SELECT {names} FROM {APP_NAME}.results.{result_table_name};
                                call {APP_NAME}.api.drop_result_table('{result_table_name}');
                            end;
                        """)

            return None

    def compile_lqp(self, model: ir.Model, task: ir.Task):
        model_txn = None
        if self._last_model != model:
            with debugging.span("compile", metamodel=model) as install_span:
                _, model_txn = self.compiler.compile(model, {"fragment_id": b"model"})
                install_span["compile_type"] = "model"
                install_span["lqp"] = lqp_print.to_string(model_txn, {"print_names": True, "print_debug": False})
                self._last_model = model

        with debugging.span("compile", metamodel=task) as compile_span:
            query = f.compute_model(f.logical([task]))
            options = {
                "wide_outputs": self.wide_outputs,
                "fragment_id": b"query",
            }
            result, final_model = self.compiler.compile_inner(query, options)
            export_info, query_txn = result
            compile_span["compile_type"] = "query"
            compile_span["lqp"] = lqp_print.to_string(query_txn, {"print_names": True, "print_debug": False})

        txn = query_txn
        if model_txn is not None:
            # Merge the two LQP transactions into one. Long term the query bits should all
            # go into a WhatIf action. But for now we just use two separate epochs.
            model_epoch = model_txn.epochs[0]
            query_epoch = query_txn.epochs[0]
            txn = lqp_ir.Transaction(
                epochs=[model_epoch, query_epoch],
                configure=construct_configure({}, None),
                meta=None,
            )

            # Revalidate now that we've joined two epochs
            validate_lqp(txn)

        txn_proto = convert_transaction(txn)
        # TODO (azreika): Should export_info be encoded as part of the txn_proto? [RAI-40312]
        return final_model, export_info, txn_proto

    # TODO (azreika): This should probably be split up into exporting and other processing. There are quite a lot of arguments here...
    def _process_results(self, task: ir.Task, final_model: ir.Model, raw_results: TransactionAsyncResponse, result_cols: Optional[list[str]], export_info: Optional[tuple], export_to: Optional[str], update: bool) -> DataFrame:
        cols, extra_cols = self._compute_cols(task, final_model)

        df, errs = result_helpers.format_results(raw_results, cols)
        self.report_errors(errs)

        # Process exports
        if export_to and not self.dry_run:
            assert cols, "No columns found in the output"
            assert isinstance(raw_results, TransactionAsyncResponse) and raw_results.transaction, "Invalid transaction result"

            if result_cols is not None:
                assert all(col in result_cols or col in extra_cols for col in cols)
            else:
                result_cols = cols
            assert result_cols

            assert export_info, "Export info should be populated if we are exporting results"
            self._export(raw_results.transaction['id'], export_info, export_to, cols, result_cols, update)

        return self._postprocess_df(self.config, df, extra_cols)

    def execute(self, model: ir.Model, task:ir.Task, result_cols:Optional[list[str]]=None, export_to:Optional[str]=None, update:bool=False) -> DataFrame:
        self.prepare_data()

        final_model, export_info, txn_proto = self.compile_lqp(model, task)

        if self.dry_run:
            return DataFrame()

        raw_results = self.resources.exec_lqp(
            self.database,
            self.engine,
            txn_proto.SerializeToString(),
            # Current strategy is to run all queries as write transactions, in order to
            # benefit from view caching. This will have to be revisited, because write
            # transactions are serialized.
            readonly=False,
            nowait_durable=True,
        )
        assert isinstance(raw_results, TransactionAsyncResponse)

        return self._process_results(task, final_model, raw_results, result_cols, export_info, export_to, update)



from abc import ABC, abstractmethod
import typing
from .glushkov import Glushkov
from .filter import (
    NodeFilter, NodeLabel, AnonymousNodeFilter, EdgeLabel, AnonymousEdgeFilter,
    parse_label
)
from relationalai.early_access.builder import Concept

# =========================================================================================
# Regular Path Queries (RPQs) with local filters
# =========================================================================================
#
# Regular Path Queries with local filters are represented as ASTs defined with this grammar.
#
# expr ::= Node([NodeLabel], [AnonymousNodeFilter])
#        | Edge(EdgeLabel, [AnonymousEdgeFilter])
#        | Concat(expr₁, expr₂)
#        | Union(expr₁, expr₂)
#        | Star(expr)
#        | Plus(expr)
#        | Optional(expr)
#        | Void
#
# Every Edge leaf of an RPQ is automatically assigned a unique position `pos` value i.e.,
# the `Edge` constructor uses a class variable `_id_counter` to generate a fresh value for
# each new instance of `Edge`. The positions are used to provide the user with clear
# feedback (error and warnings) that refers to specific places in the input RPQ.
#
# An important requirement is that an RPQ does not use two Edge leaves with the same
# position.  With a natural use of the RPQ constructors, this requirement is not violated
# unless a subexpression is reused. Consequently, the `clone` method is provided to create a
# copy of an RPQ subexpression with fresh edge positions. For further safety, we ensure that
# no two Edges with the same position are present, by constructing the set of positions of
# an expression. If during the construction of a complex RPQ (`Concat` or `Union`) we find
# repeating positions, we clone one of the expressions to ensure that the positions are
# unique in the whole RPQ.
# =========================================================================================

#------------------------------------------------------------------------------------------
# Auxiliary function for repeating `expr{n:m}` used to implement slicing operator of RPQ
#------------------------------------------------------------------------------------------


#-----------------------------------------------------------------------------
# Parsing RPQ elements specified with strings (node and edge labels).
#-----------------------------------------------------------------------------


class RPQ(ABC):
    _id_counter = 0

    @classmethod
    def next_id(cls) -> int:
        cls._id_counter += 1
        return cls._id_counter

    def __init__(self, positions: typing.FrozenSet[int] = frozenset()):
        self._positions = positions

    @property
    def positions(self) -> typing.FrozenSet[int]:
        return self._positions

    @abstractmethod
    def clone(self) -> 'RPQ':
        pass

    @abstractmethod
    def glushkov(self) -> Glushkov:
        pass

    @abstractmethod
    def pprint(self, ind) -> str:
        pass

    def __str__(self):
        return self.pprint("")

    @abstractmethod
    def to_dict(self) -> dict:
        pass

    def __len__(self) -> int:
        return len(self.positions)

    # expr[n:m]
    def __getitem__(self, key):
        if isinstance(key, int):
            return _repeat(self, key, key)
        elif isinstance(key, slice):
            return _repeat(self, key.start, key.stop)
        else:
            return NotImplemented

    # expr₁ | expr₂
    def __or__(self, other):
        if isinstance(self, Void):
            return other
        elif isinstance(other, Void):
            return self
        else:
            return Union(self, other)

    # expr₁ ⋅ expr₂
    def __mul__(self, other):
        return Concat(self, other)

# ∅
class Void(RPQ):
    def __init__(self):
        super().__init__()

    def clone(self):
        return Void()

    def glushkov(self):
        return Glushkov.void()

    def pprint(self, ind):
        return ind + "⋅ ∅"

    def to_dict(self):
        return {"type": "Void"}

# ε(label {filter})
class Node(RPQ):
    def __init__(self,
                 label: typing.Optional[NodeLabel]=None,
                 filter: typing.Optional[AnonymousNodeFilter]=None):
        super().__init__()
        self.label = label
        self.filter = filter

    def clone(self):
        return Node(self.label, self.filter)

    def glushkov(self):
        filters: typing.List[NodeFilter] = []
        if self.label is not None:
            filters.append(self.label)
        if self.filter is not None:
            filters.append(self.filter)
        return Glushkov.node_transition(*filters)

    def pprint(self, ind):
        res = ind + "⋅ Node("
        if self.label is not None:
            res += f"{self.label}"
        if self.filter is not None:
            res += " {" if self.label is not None else "{"
            res += f"{self.filter}"
            res += "}"
        res += ")"
        return res

    def to_dict(self) -> dict:
        d = {"type": "Node"}
        if self.label is not None:
            d["label"] = self.label.to_dict() # type: ignore
        if self.filter is not None:
            d["filter"] = self.filter.to_dict() # type: ignore
        return d

# -[label {filter}]->
class Edge(RPQ):
    def __init__(self,
                 label: EdgeLabel,
                 filter: typing.Optional[AnonymousEdgeFilter] = None):
        self.pos = self.next_id()
        super().__init__(frozenset([self.pos]))
        self.label = label
        self.filter = filter

    def clone(self):
        return Edge(self.label, self.filter)

    def glushkov(self):
        return Glushkov.edge_transition(self.pos, self.label, self.filter)

    # we avoid using `[...]` which can be mistaken for control sequence by rich-text loggers
    def pprint(self, ind):
        res = ind + "⋅ "
        filter_str = ""
        if self.filter is not None:
            filter_str += f" {{ {self.filter} }}"
        if self.label.direction == "forward":
            res += f"-⟨{self.label.label}{filter_str}⟩→"
        else:
            res += f"←⟨{self.label.label}{filter_str}⟩-"
        res += f" pos={self.pos} "
        return res

    def to_dict(self) -> dict:
        return {
            "type": "Edge",
            "pos": self.pos,
            "label": self.label.to_dict() # type: ignore
        }

# expr₁ ⋅ expr₂
class Concat(RPQ):
    def __init__(self, expr1:RPQ, expr2:RPQ):
        if not expr1.positions.isdisjoint(expr2.positions):
            expr2 = expr2.clone()
        super().__init__(expr1.positions | expr2.positions)
        self.expr1 = expr1
        self.expr2 = expr2

    def clone(self):
        return Concat(self.expr1.clone(), self.expr2.clone())

    def glushkov(self):
        g = self.expr1.glushkov()
        g *= self.expr2.glushkov()
        return g

    def pprint(self, ind):
        res = ind + "⋅ Concat\n"
        res += self.expr1.pprint(ind + "   ") + "\n"
        res += self.expr2.pprint(ind + "   ")
        return res

    def to_dict(self) -> dict:
        return {
            "type": "Concat",
            "expr1": self.expr1.to_dict(), # type: ignore
            "expr2": self.expr2.to_dict()  # type: ignore
        }

# expr ₁ | expr₂
class Union(RPQ):
    def __init__(self, expr1:RPQ, expr2:RPQ):
        if not expr1.positions.isdisjoint(expr2.positions):
            expr2 = expr2.clone()
        super().__init__(expr1.positions | expr2.positions)
        self.expr1 = expr1
        self.expr2 = expr2

    def clone(self):
        return Union(self.expr1.clone(), self.expr2.clone())

    def glushkov(self):
        g = self.expr1.glushkov()
        g |= self.expr2.glushkov()
        return g

    def pprint(self, ind):
        res = ind + "⋅ Union\n"
        res += self.expr1.pprint(ind + "   ") + "\n"
        res += self.expr2.pprint(ind + "   ")
        return res

    def to_dict(self) -> dict:
        return {
            "type": "Union",
            "expr1": self.expr1.to_dict(), # type: ignore
            "expr2": self.expr2.to_dict()  # type: ignore
        }

# expr*
class Star(RPQ):
    def __init__(self, expr:RPQ):
        super().__init__(expr.positions)
        self.expr = expr

    def clone(self):
        return Star(self.expr.clone())

    def glushkov(self):
        return self.expr.glushkov().star()

    def pprint(self, ind):
        return ind + "⋅ Star\n" + self.expr.pprint(ind + '   ')

    def to_dict(self) -> dict:
        return {
            "type": "Star",
            "expr": self.expr.to_dict() # type: ignore
        }

# expr+
class Plus(RPQ):
    def __init__(self, expr:RPQ):
        super().__init__(expr.positions)
        self.expr = expr

    def clone(self):
        return Plus(self.expr.clone())

    def glushkov(self):
        return self.expr.glushkov().plus()

    def pprint(self, ind):
        return ind + "⋅ Plus\n" + self.expr.pprint(ind + '   ')

    def to_dict(self) -> dict:
        return {
            "type": "Plus",
            "expr": self.expr.to_dict() # type: ignore
        }

# expr?
class Optional(RPQ):
    def __init__(self, expr:RPQ):
        super().__init__(expr.positions)
        self.expr = expr

    def clone(self):
        return Optional(self.expr.clone())

    def glushkov(self):
        return self.expr.glushkov().optional()

    def pprint(self, ind):
        return ind + "⋅ Optional\n" + self.expr.pprint(ind + '   ')

    def to_dict(self) -> dict:
        return {
            "type": "Optional",
            "expr": self.expr.to_dict() # type: ignore
        }



# allowed types for segments in RPQ constructors:
Segment = typing.Union[str, Concept, RPQ]


def _process_segment(s: Segment) -> RPQ:

    if isinstance(s, str):
        label = parse_label(s)
        if isinstance(label, NodeLabel):
            return Node(label)
        else:
            return Edge(label)
    elif isinstance(s, Concept):
        return Node(NodeLabel(s._name))
    elif isinstance(s, RPQ):
        return s
    else:
        raise Exception(f"Incorrect type `{type(s)}` of segment `{s}`")


def _repeat(pattern: 'RPQ', n:typing.Union[int,None], m:typing.Union[int,None]) -> 'RPQ':
    n = n if n is not None else 0
    assert n >= 0, f"Negative start range [{n}:{m}]"
    assert (m is None or m >= n), f"Invalid repeat range [{n}:{m}]"
    if m is None:
        if n == 0:
            return Star(pattern)
        elif n == 1:
            return Plus(pattern)
        else:
            result = pattern.clone()
            for _ in range(n-2):
                result = Concat(result, pattern.clone())
            return Concat(result, Plus(pattern))
    else:
        if n == 0:
            if m == 0:
                return Node()
            else:
                result = Optional(pattern)
                for _ in range(m-1):
                    result = Concat(result, Optional(pattern.clone()))
                return result
        else:
            result = pattern
            for _ in range(n-1):
                result = Concat(result, pattern.clone())
            for _ in range(m-n):
                result = Concat(result, Optional(pattern.clone()))
            return result

# Explicitly export RPQ and other relevant classes
__all__ = ["RPQ", "Node", "Edge", "Concat", "Union", "Star"]


from __future__ import annotations
from typing import Any

from .. import builder as b
from .std import _make_expr

def rule_reasoner_semantic_variable_order(*args: Any) -> Any:
    return _make_expr("rule_reasoner_sem_vo", b.TupleArg(args))

def rule_reasoner_physical_variable_order(*args: Any) -> Any:
    return _make_expr("rule_reasoner_phys_vo", b.TupleArg(args))



from __future__ import annotations

from pandas import DataFrame
from typing import Any, Union, Tuple

from relationalai.clients.config import Config
from relationalai.early_access.metamodel import Model, Task, ir
from relationalai.early_access.metamodel.visitor import collect_by_type
from .util import NameCache

import rich

class Executor():
    """ Interface for an object that can execute the program specified by a model. """
    def execute(self, model: Model, task:Task) -> Union[DataFrame, Any]:
        raise NotImplementedError(f"execute: {self}")

    def _compute_cols(self, task: ir.Task, final_model: ir.Model|None) -> Tuple[list[str], list[str]]:
        cols = []
        extra_cols = []
        # we assume only queries have outputs
        original_outputs = collect_by_type(ir.Output, task) if task else None
        outputs = collect_by_type(ir.Output, final_model) if final_model else None
        # there are some outputs, and they all have keys
        if original_outputs and outputs and not all(not out.keys for out in outputs):
            assert len(original_outputs) == 1
            original_output = original_outputs[0]
            original_cols = []
            original_cols_val = []
            for alias, val in original_output.aliases:
                if not alias:
                    continue
                original_cols.append(alias)
                original_cols_val.append(val)

            keys = outputs[0].keys
            assert keys
            for out in outputs:
                assert out.keys == keys, "outputs with different key sets in the same query"

            extra_cols = []
            name_cache = NameCache(start_from_one=True)
            for key in keys:
                if isinstance(key, ir.Var) and key not in original_cols_val:
                    extra_cols.append(f"hidden_{name_cache.get_name(key.id, key.name)}")
            cols = original_cols + extra_cols
        elif outputs:
            cols = [alias for alias, _ in outputs[-1].aliases if alias]

        return cols, extra_cols

    def _postprocess_df(self, config: Config, df: DataFrame, extra_cols: list[str]) -> DataFrame:
        if bool(config.get("compiler.debug_hidden_keys", False)):
            rich.print("[blue]DataFrame with extra hidden columns:[/blue]")
            rich.print(f"[black]{df}[/black]")
        if not df.empty:
            for col in extra_cols:
                df = df.drop(col, axis=1)
        return df



from __future__ import annotations

import duckdb
from pandas import DataFrame
from typing import Any, Union

from relationalai.semantics.sql import Compiler
from relationalai.semantics.sql.executor.result_helpers import format_duckdb_columns
from relationalai.early_access.metamodel import ir, executor as e

class DuckDBExecutor(e.Executor):

    def __init__(self, skip_denormalization: bool = False) -> None:
        super().__init__()
        self.compiler = Compiler(skip_denormalization)

    def execute(self, model: ir.Model, task: ir.Task) -> Union[DataFrame, Any]:
        """ Execute the SQL query directly. """
        connection = duckdb.connect()
        try:
            sql, _ = self.compiler.compile(model, {"is_duck_db": True})
            arrow_table = connection.query(sql).fetch_arrow_table()
            return format_duckdb_columns(arrow_table.to_pandas(), arrow_table.schema)
        finally:
            connection.close()
